@@ -1,4 +1,4 @@
-# Copyright 2004-2014 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -23,6 +23,7 @@
 # size-based caching and constructing images from operations (like
 # cropping and scaling).
 
+from __future__ import print_function
 import renpy.display
 
 import math
@@ -30,28 +31,50 @@ import zipfile
 import cStringIO
 import threading
 import time
+import io
 
 
 # This is an entry in the image cache.
 class CacheEntry(object):
 
-    def __init__(self, what, surf):
+    def __init__(self, what, surf, bounds):
 
         # The object that is being cached (which needs to be
         # hashable and comparable).
         self.what = what
 
-        # The pygame surface corresponding to the cached object.
+        # The pygame surface corresponding to the cached object. This may be
+        # None if we've tossed the surface.
         self.surf = surf
 
-        # The size of this image.
-        w, h = surf.get_size()
-        self.size = w * h
+        # The sizes of surf.
+        self.width, self.height = surf.get_size()
+
+        # The texture corresponding to the visible area of the cached object.
+        # This may be None if no texture has been loaded.
+        self.texture = None
+
+        # The bounds of the texture within the width and height.
+        self.bounds = bounds
 
         # The time when this cache entry was last used.
         self.time = 0
 
+    def size(self):
+        rv = 0
+
+        if self.surf is not None:
+            rv += self.width * self.height
+
+        if self.texture is not None:
+            rv += self.bounds[2] * self.bounds[3]
+
+        return rv
+
+
 # This is the singleton image cache.
+
+
 class Cache(object):
 
     def __init__(self):
@@ -69,12 +92,6 @@ class Cache(object):
         # False if this is not the first preload in this tick.
         self.first_preload_in_tick = True
 
-        # The total size of the current generation of images.
-        self.size_of_current_generation = 0
-
-        # The total size of everything in the cache.
-        self.total_cache_size = 0
-
         # A lock that must be held when updating the cache.
         self.lock = threading.Condition()
 
@@ -111,6 +128,32 @@ class Cache(object):
         # This is only updated when config.developer is True.
         self.load_log = [ ]
 
+    def get_total_size(self):
+        """
+        Returns the total size of the surfaces and textures that make up the
+        cache, in pixels.
+        """
+
+        rv = sum(i.size() for i in self.cache.values())
+
+#         print("Total cache size: {:.1f}/{:.1f} MB (Textures {:.1f} MB)".format(
+#             4.0 * rv / 1024 / 1024,
+#             4.0 * self.cache_limit / 1024 / 1024,
+#             1.0 * renpy.exports.get_texture_size()[0] / 1024 / 1024,
+#             ))
+
+        return rv
+
+    def get_current_size(self, generations):
+        """
+        Returns the size of the most recent `generation` generations of
+        the cache. (1 is the current, 2 is the current and one before).
+        """
+
+        start = self.time - generations
+
+        rv = sum(i.size() for i in self.cache.values() if i.time > start)
+        return rv
 
     def init(self):
         """
@@ -118,9 +161,12 @@ class Cache(object):
         by the game-maker.
         """
 
-        self.cache_limit = renpy.config.image_cache_size * renpy.config.screen_width * renpy.config.screen_height
+        if renpy.config.image_cache_size is not None:
+            self.cache_limit = 2 * renpy.config.image_cache_size * renpy.config.screen_width * renpy.config.screen_height
+        else:
+            self.cache_limit = int(renpy.config.image_cache_size_mb * 1024 * 1024 // 4)
 
-    def quit(self): #@ReservedAssignment
+    def quit(self):  # @ReservedAssignment
         if not self.preload_thread.isAlive():
             return
 
@@ -132,7 +178,6 @@ class Cache(object):
 
         self.clear()
 
-
     # Clears out the cache.
     def clear(self):
 
@@ -142,8 +187,6 @@ class Cache(object):
         self.pin_cache = { }
         self.cache = { }
         self.first_preload_in_tick = True
-        self.size_of_current_generation = 0
-        self.total_cache_size = 0
 
         self.added.clear()
 
@@ -157,7 +200,6 @@ class Cache(object):
             self.time += 1
             self.preloads = [ ]
             self.first_preload_in_tick = True
-            self.size_of_current_generation = 0
             self.added.clear()
 
         if renpy.config.debug_image_cache:
@@ -170,12 +212,16 @@ class Cache(object):
     def end_tick(self):
         self.preloads = [ ]
 
-
     # This returns the pygame surface corresponding to the provided
     # image. It also takes care of updating the age of images in the
     # cache to be current, and maintaining the size of the current
     # generation of images.
-    def get(self, image, predict=False):
+    def get(self, image, predict=False, texture=False, render=False):
+
+        if render:
+            texture = True
+
+        optimize_bounds = renpy.config.optimize_texture_bounds
 
         if not isinstance(image, ImageBase):
             raise Exception("Expected an image of some sort, but got" + str(image) + ".")
@@ -188,6 +234,25 @@ class Cache(object):
         # First try to grab the image out of the cache without locking it.
         ce = self.cache.get(image, None)
 
+        if ce is not None:
+
+            ce.time = self.time
+
+            if texture and (ce.texture is not None):
+
+                if predict:
+                    return None
+
+                if render:
+                    rv = renpy.display.render.Render(ce.width, ce.height)
+                    rv.blit(ce.texture, ce.bounds[:2])
+                    return rv
+                else:
+                    return ce.texture
+
+            if ce.surf is None:
+                ce = None
+
         # Otherwise, we load the image ourselves.
         if ce is None:
 
@@ -195,18 +260,28 @@ class Cache(object):
                 if image in self.pin_cache:
                     surf = self.pin_cache[image]
                 else:
-                    surf = image.load()
+
+                    if not predict:
+                        with renpy.game.ExceptionInfo("While loading %r:", image):
+                            surf = image.load()
+                    else:
+                        surf = image.load()
 
             except:
                 raise
 
+            w, h = surf.get_size()
+
+            if optimize_bounds:
+                bounds = tuple(surf.get_bounding_rect())
+                w = bounds[2]
+                h = bounds[3]
+            else:
+                bounds = (0, 0, w, h)
+
             with self.lock:
 
-                ce = CacheEntry(image, surf)
-
-                if image not in self.cache:
-                    self.total_cache_size += ce.size
-
+                ce = CacheEntry(image, surf, bounds)
                 self.cache[image] = ce
 
                 # Indicate that this surface had changed.
@@ -214,34 +289,61 @@ class Cache(object):
 
                 if renpy.config.debug_image_cache:
                     if predict:
-                        renpy.display.ic_log.write("Added %r (%.02f%%)", ce.what, 100.0 * self.total_cache_size / self.cache_limit)
+                        renpy.display.ic_log.write("Added %r (%.02f%%)", ce.what, 100.0 * self.get_total_size() / self.cache_limit)
                     else:
                         renpy.display.ic_log.write("Total Miss %r", ce.what)
 
-                renpy.display.draw.load_texture(ce.surf)
+        # Move it into the current generation.
 
+        ce.time = self.time
 
-        # Move it into the current generation. This isn't protected by
-        # a lock, so in certain circumstances we could have an
-        # inaccurate size - but that will be cured at the end of the
-        # current generation.
+        # Load the texture.
 
-        if ce.time != self.time:
-            ce.time = self.time
-            self.size_of_current_generation += ce.size
+        if texture:
+
+            if ce.texture is None:
+
+                texsurf = ce.surf
+
+                if ce.bounds != (0, 0, ce.width, ce.height):
+                    texsurf = ce.surf.subsurface(ce.bounds)
+                    renpy.display.render.mutated_surface(texsurf)
+
+                ce.texture = renpy.display.draw.load_texture(texsurf)
+
+            if not predict:
+                if render:
+                    rv = renpy.display.render.Render(ce.width, ce.height)
+                    rv.blit(ce.texture, ce.bounds[:2])
+                else:
+                    rv = ce.texture
+            else:
+                rv = None
+
+        else:
+            rv = ce.surf
+
+        if not renpy.config.cache_surfaces:
+
+            if ce.surf is not None:
+                renpy.display.draw.mutated_surface(ce.surf)
+
+            ce.surf = None
+
+        if (ce.surf is None) and (ce.texture is None):
+            with self.lock:
+                self.kill(ce)
 
         # Done... return the surface.
-        return ce.surf
-
+        return rv
 
     # This kills off a given cache entry.
     def kill(self, ce):
 
-        # Should never happen... but...
-        if ce.time == self.time:
-            self.size_of_current_generation -= ce.size
+        # Let the texture cache know we're not needed.
+        if ce.surf is not None:
+            renpy.display.draw.mutated_surface(ce.surf)
 
-        self.total_cache_size -= ce.size
         del self.cache[ce.what]
 
         if renpy.config.debug_image_cache:
@@ -255,7 +357,7 @@ class Cache(object):
         """
 
         # If we're within the limit, return.
-        if self.total_cache_size <= self.cache_limit:
+        if self.get_total_size() <= self.cache_limit:
             return True
 
         # If we're outside the cache limit, we need to go and start
@@ -272,11 +374,26 @@ class Cache(object):
             self.kill(ce)
 
             # If we're in the limit, we're done.
-            if self.total_cache_size <= self.cache_limit:
+            if self.get_total_size() <= self.cache_limit:
                 break
 
         return True
 
+    def preload_texture(self, im):
+        """
+        Preloads `im` into the cache, and loads the corresponding texture
+        into the GPU.
+        """
+
+        self.get(im, predict=True, texture=True)
+
+    def get_texture(self, im):
+        """
+        Gets `im` as a texture. Used when prediction is being used to load
+        the actual image.
+        """
+
+        self.get(im, texture=True)
 
     # Called to report that a given image would like to be preloaded.
     def preload_image(self, im):
@@ -291,8 +408,10 @@ class Cache(object):
 
             self.added.add(im)
 
-            if im in self.cache:
-                self.get(im)
+            ce = self.cache.get(im, None)
+
+            if ce and ce.texture:
+                ce.time = self.time
                 in_cache = True
             else:
                 self.preloads.append(im)
@@ -306,7 +425,6 @@ class Cache(object):
         if in_cache and renpy.config.debug_image_cache:
             renpy.display.ic_log.write("Kept %r", im)
 
-
     def start_prediction(self):
         """
         Called at the start of prediction, to ensure the thread runs
@@ -324,70 +442,73 @@ class Cache(object):
             self.preload_lock.wait()
             self.preload_lock.release()
 
-            while self.preloads and self.keep_preloading:
+            self.preload_thread_pass()
 
-                # If the size of the current generation is bigger than the
-                # total cache size, stop preloading.
-                with self.lock:
+    def preload_thread_pass(self):
 
-                    # If the cache is overfull, clean it out.
-                    if not self.cleanout():
-
-                        if renpy.config.debug_image_cache:
-                            for i in self.preloads:
-                                renpy.display.ic_log.write("Overfull %r", i)
-
-                        self.preloads = [ ]
-
-                        break
-
-                try:
-                    image = self.preloads.pop(0)
-
-                    if image not in self.preload_blacklist:
-                        try:
-                            self.get(image, True)
-                        except:
-                            self.preload_blacklist.add(image)
-                except:
-                    pass
+        while self.preloads and self.keep_preloading:
 
+            # If the size of the current generation is bigger than the
+            # total cache size, stop preloading.
             with self.lock:
-                self.cleanout()
 
-            # If we have time, preload pinned images.
-            if self.keep_preloading and not renpy.game.less_memory:
+                # If the cache is overfull, clean it out.
+                if not self.cleanout():
 
-                workset = set(renpy.store._cache_pin_set)
+                    if renpy.config.debug_image_cache:
+                        for i in self.preloads:
+                            renpy.display.ic_log.write("Overfull %r", i)
 
-                # Remove things that are not in the workset from the pin cache,
-                # and remove things that are in the workset from pin cache.
-                for i in self.pin_cache.keys():
+                    self.preloads = [ ]
 
-                    if i in workset:
-                        workset.remove(i)
-                    else:
-                        surf = self.pin_cache[i]
+                    break
 
-                        del self.pin_cache[i]
-
-
-                # For each image in the worklist...
-                for image in workset:
-
-                    if image in self.preload_blacklist:
-                        continue
-
-                    # If we have normal preloads, break out.
-                    if self.preloads:
-                        break
+            try:
+                image = self.preloads.pop(0)
 
+                if image not in self.preload_blacklist:
                     try:
-                        surf = image.load()
-                        self.pin_cache[image] = surf
-                        renpy.display.draw.load_texture(surf)
+                        self.preload_texture(image)
                     except:
                         self.preload_blacklist.add(image)
+            except:
+                pass
+
+        with self.lock:
+            self.cleanout()
+
+        # If we have time, preload pinned images.
+        if self.keep_preloading and not renpy.game.less_memory:
+
+            workset = set(renpy.store._cache_pin_set)
+
+            # Remove things that are not in the workset from the pin cache,
+            # and remove things that are in the workset from pin cache.
+            for i in self.pin_cache.keys():
+
+                if i in workset:
+                    workset.remove(i)
+                else:
+                    surf = self.pin_cache[i]
+
+                    del self.pin_cache[i]
+
+            # For each image in the worklist...
+            for image in workset:
+
+                if image in self.preload_blacklist:
+                    continue
+
+                # If we have normal preloads, break out.
+                if self.preloads:
+                    break
+
+                try:
+                    surf = image.load()
+                    self.pin_cache[image] = surf
+                    renpy.display.draw.load_texture(surf)
+                except:
+                    self.preload_blacklist.add(image)
 
     def add_load_log(self, filename):
 
@@ -402,16 +523,15 @@ class Cache(object):
             self.load_log.pop()
 
 
-
 # The cache object.
 cache = Cache()
 
+
 def free_memory():
     """
     Frees some memory.
     """
 
-    renpy.display.draw.free_memory()
     cache.clear()
 
 
@@ -437,7 +557,6 @@ class ImageBase(renpy.display.core.Displayable):
         super(ImageBase, self).__init__(**properties)
         self.identity = (type(self).__name__, ) + args
 
-
     def __hash__(self):
         return hash(self.identity)
 
@@ -461,14 +580,7 @@ class ImageBase(renpy.display.core.Displayable):
         raise Exception("load method not implemented.")
 
     def render(self, w, h, st, at):
-
-        im = cache.get(self)
-        texture = renpy.display.draw.load_texture(im)
-
-        w, h = im.get_size()
-        rv = renpy.display.render.Render(w, h)
-        rv.blit(texture, (0, 0))
-        return rv
+        return cache.get(self, render=True)
 
     def predict_one(self):
         renpy.display.predict.image(self)
@@ -481,6 +593,7 @@ class ImageBase(renpy.display.core.Displayable):
 
         return [ ]
 
+
 class Image(ImageBase):
     """
     This image manipulator loads an image from a file.
@@ -494,8 +607,14 @@ class Image(ImageBase):
         super(Image, self).__init__(filename, **properties)
         self.filename = filename
 
-    def get_mtime(self):
-        return renpy.loader.get_mtime(self.filename)
+    def __unicode__(self):
+        if len(self.filename) < 20:
+            return u"Image %r" % self.filename
+        else:
+            return u"Image \u2026%s" % self.filename[-20:]
+
+    def get_hash(self):
+        return renpy.loader.get_hash(self.filename)
 
     def load(self, unscaled=False):
 
@@ -510,7 +629,7 @@ class Image(ImageBase):
 
             return surf
 
-        except Exception, e:
+        except Exception as e:
 
             if renpy.config.missing_image_callback:
                 im = renpy.config.missing_image_callback(self.filename)
@@ -533,6 +652,36 @@ class Image(ImageBase):
 
             return [ self.filename ]
 
+
+class Data(ImageBase):
+    """
+    :doc: im_im
+
+    This image manipulator loads an image from binary data.
+
+    `data`
+        A string of bytes, giving the compressed image data in a standard
+        file format.
+
+    `filename`
+        A "filename" associated with the image. This is used to provide a
+        hint to Ren'Py about the format of `data`. (It's not actually
+        loaded from disk.)
+    """
+
+    def __init__(self, data, filename, **properties):
+        super(Data, self).__init__(data, filename, **properties)
+        self.data = data
+        self.filename = filename
+
+    def __unicode__(self):
+        return u"im.Data(%r)" % self.filename
+
+    def load(self):
+        f = io.BytesIO(self.data)
+        return renpy.display.pgrender.load_image(f, self.filename)
+
+
 class ZipFileImage(ImageBase):
 
     def __init__(self, zipfilename, filename, mtime=0, **properties):
@@ -552,16 +701,13 @@ class ZipFileImage(ImageBase):
         except:
             return renpy.display.pgrender.surface((2, 2), True)
 
-
-
     def predict_files(self):
         return [ ]
 
 
-
 class Composite(ImageBase):
     """
-    :doc: im_im
+    :undocumented:
 
     This image manipulator composites multiple images together to
     form a single image.
@@ -597,8 +743,13 @@ class Composite(ImageBase):
         self.positions = args[0::2]
         self.images = [ image(i) for i in args[1::2] ]
 
-    def get_mtime(self):
-        return min(i.get_mtime() for i in self.images)
+    def get_hash(self):
+        rv = 0
+
+        for i in self.images:
+            rv += i.get_hash()
+
+        return rv
 
     def load(self):
 
@@ -623,9 +774,10 @@ class Composite(ImageBase):
 
         return rv
 
+
 class Scale(ImageBase):
     """
-    :doc: im_im
+    :undocumented:
 
     An image manipulator that scales `im` (an image manipulator) to
     `width` and `height`.
@@ -648,8 +800,8 @@ class Scale(ImageBase):
         self.height = int(height)
         self.bilinear = bilinear
 
-    def get_mtime(self):
-        return self.image.get_mtime()
+    def get_hash(self):
+        return self.image.get_hash()
 
     def load(self):
 
@@ -673,13 +825,14 @@ class Scale(ImageBase):
     def predict_files(self):
         return self.image.predict_files()
 
+
 class FactorScale(ImageBase):
     """
     :doc: im_im
 
     An image manipulator that scales `im` (a second image manipulator)
     to `width` times its original `width`, and `height` times its
-    original height. If `height` is ommitted, it defaults to `width`.
+    original height. If `height` is omitted, it defaults to `width`.
 
     If `bilinear` is true, then bilinear interpolation is used for
     the scaling. Otherwise, nearest neighbor interpolation is used.
@@ -689,7 +842,6 @@ class FactorScale(ImageBase):
         image logo doubled = im.FactorScale("logo.png", 1.5)
     """
 
-
     def __init__(self, im, width, height=None, bilinear=True, **properties):
 
         if height is None:
@@ -703,8 +855,8 @@ class FactorScale(ImageBase):
         self.height = height
         self.bilinear = bilinear
 
-    def get_mtime(self):
-        return self.image.get_mtime()
+    def get_hash(self):
+        return self.image.get_hash()
 
     def load(self):
 
@@ -759,9 +911,8 @@ class Flip(ImageBase):
         self.horizontal = horizontal
         self.vertical = vertical
 
-
-    def get_mtime(self):
-        return self.image.get_mtime()
+    def get_hash(self):
+        return self.image.get_hash()
 
     def load(self):
 
@@ -775,12 +926,10 @@ class Flip(ImageBase):
 
         return rv
 
-
     def predict_files(self):
         return self.image.predict_files()
 
 
-
 class Rotozoom(ImageBase):
     """
     This is an image manipulator that is a smooth rotation and zoom of another image manipulator.
@@ -804,8 +953,8 @@ class Rotozoom(ImageBase):
         self.angle = angle
         self.zoom = zoom
 
-    def get_mtime(self):
-        return self.image.get_mtime()
+    def get_hash(self):
+        return self.image.get_hash()
 
     def load(self):
 
@@ -823,7 +972,6 @@ class Rotozoom(ImageBase):
         return self.image.predict_files()
 
 
-
 class Crop(ImageBase):
     """
     :doc: im_im
@@ -852,8 +1000,8 @@ class Crop(ImageBase):
         self.w = w
         self.h = h
 
-    def get_mtime(self):
-        return self.image.get_mtime()
+    def get_hash(self):
+        return self.image.get_hash()
 
     def load(self):
         return cache.get(self.image).subsurface((self.x, self.y,
@@ -887,8 +1035,10 @@ def ramp(start, end):
 
     return rv
 
+
 identity = ramp(0, 255)
 
+
 class Map(ImageBase):
     """
     This adjusts the colors of the image that is its child. It takes
@@ -912,8 +1062,8 @@ class Map(ImageBase):
 
         self.force_alpha = force_alpha
 
-    def get_mtime(self):
-        return self.image.get_mtime()
+    def get_hash(self):
+        return self.image.get_hash()
 
     def load(self):
 
@@ -929,6 +1079,7 @@ class Map(ImageBase):
     def predict_files(self):
         return self.image.predict_files()
 
+
 class Twocolor(ImageBase):
     """
     This takes as arguments two colors, white and black. The image is
@@ -954,8 +1105,8 @@ class Twocolor(ImageBase):
 
         self.force_alpha = force_alpha
 
-    def get_mtime(self):
-        return self.image.get_mtime()
+    def get_hash(self):
+        return self.image.get_hash()
 
     def load(self):
 
@@ -994,8 +1145,8 @@ class Recolor(ImageBase):
 
         self.force_alpha = force_alpha
 
-    def get_mtime(self):
-        return self.image.get_mtime()
+    def get_hash(self):
+        return self.image.get_hash()
 
     def load(self):
 
@@ -1011,6 +1162,50 @@ class Recolor(ImageBase):
     def predict_files(self):
         return self.image.predict_files()
 
+
+class Blur(ImageBase):
+    """
+    :doc: im_im
+
+    An image manipulator that blurs the image manipulator `im` using
+    an elliptical kernel described by `xrad` and optionally `yrad`.
+
+    If `yrad` is None, it will take the value of `xrad` resulting in
+    a circular kernel being used.
+
+    ::
+
+        image logo blurred = im.Blur("logo.png", 1.5)
+    """
+
+    def __init__(self, im, xrad, yrad=None, **properties):
+
+        im = image(im)
+
+        super(Blur, self).__init__(im, xrad, yrad, **properties)
+
+        self.image = im
+        self.rx = xrad
+        self.ry = xrad if yrad is None else yrad
+
+    def get_hash(self):
+        return self.image.get_hash()
+
+    def load(self):
+
+        surf = cache.get(self.image)
+
+        ws = renpy.display.pgrender.surface(surf.get_size(), True)
+        rv = renpy.display.pgrender.surface(surf.get_size(), True)
+
+        renpy.display.module.blur(surf, ws, rv, self.rx, self.ry)
+
+        return rv
+
+    def predict_files(self):
+        return self.image.predict_files()
+
+
 class MatrixColor(ImageBase):
     """
     :doc: im_matrixcolor
@@ -1056,8 +1251,8 @@ class MatrixColor(ImageBase):
         self.image = im
         self.matrix = matrix
 
-    def get_mtime(self):
-        return self.image.get_mtime()
+    def get_hash(self):
+        return self.image.get_hash()
 
     def load(self):
 
@@ -1072,6 +1267,7 @@ class MatrixColor(ImageBase):
     def predict_files(self):
         return self.image.predict_files()
 
+
 class matrix(tuple):
     """
     :doc: im_matrixcolor
@@ -1108,6 +1304,10 @@ class matrix(tuple):
             a = matrix(a)
 
         if not isinstance(b, matrix):
+
+            if isinstance(b, renpy.easy.Color):
+                return NotImplemented
+
             b = matrix(b)
 
         result = [ 0 ] * 25
@@ -1130,7 +1330,6 @@ class matrix(tuple):
                 o[0]*self[15] + o[1]*self[16] + o[2]*self[17] + o[3]*self[18] + self[19],
                 1)
 
-
     def __add__(self, other):
         if isinstance(other, (int, float)):
             other = float(other)
@@ -1167,7 +1366,6 @@ im.matrix(%f, %f, %f, %f, %f.
           %f, %f, %f, %f, %f,
           %f, %f, %f, %f, %f)""" % self
 
-
     @staticmethod
     def identity():
         """
@@ -1182,6 +1380,7 @@ im.matrix(%f, %f, %f, %f, %f.
                       0, 1, 0, 0, 0,
                       0, 0, 1, 0, 0,
                       0, 0, 0, 1, 0)
+
     @staticmethod
     def saturation(level, desat=(0.2126, 0.7152, 0.0722)):
         """
@@ -1327,11 +1526,11 @@ im.matrix(%f, %f, %f, %f, %f.
         lumG = 0.715
         lumB = 0.072
         return matrix(
-            lumR+cosVal*(1-lumR)+sinVal*(-lumR),lumG+cosVal*(-lumG)+sinVal*(-lumG),lumB+cosVal*(-lumB)+sinVal*(1-lumB),0,0,
-            lumR+cosVal*(-lumR)+sinVal*(0.143),lumG+cosVal*(1-lumG)+sinVal*(0.140),lumB+cosVal*(-lumB)+sinVal*(-0.283),0,0,
-            lumR+cosVal*(-lumR)+sinVal*(-(1-lumR)),lumG+cosVal*(-lumG)+sinVal*(lumG),lumB+cosVal*(1-lumB)+sinVal*(lumB),0,0,
-            0,0,0,1,0,
-            0,0,0,0,1
+            lumR+cosVal*(1-lumR)+sinVal*(-lumR), lumG+cosVal*(-lumG)+sinVal*(-lumG), lumB+cosVal*(-lumB)+sinVal*(1-lumB), 0, 0,
+            lumR+cosVal*(-lumR)+sinVal*(0.143), lumG+cosVal*(1-lumG)+sinVal*(0.140), lumB+cosVal*(-lumB)+sinVal*(-0.283), 0, 0,
+            lumR+cosVal*(-lumR)+sinVal*(-(1-lumR)), lumG+cosVal*(-lumG)+sinVal*(lumG), lumB+cosVal*(1-lumB)+sinVal*(lumB), 0, 0,
+            0, 0, 0, 1, 0,
+            0, 0, 0, 0, 1
             )
 
     @staticmethod
@@ -1342,7 +1541,7 @@ im.matrix(%f, %f, %f, %f, %f.
 
         Returns an im.matrix that colorizes a black and white image.
         `black_color` and `white_color` are Ren'Py style colors, so
-        they may be specfied as strings or tuples of (0-255) color
+        they may be specified as strings or tuples of (0-255) color
         values. ::
 
             # This makes black colors red, and white colors blue.
@@ -1368,13 +1567,12 @@ im.matrix(%f, %f, %f, %f, %f.
                       0, 0, 0, 1, 0)
 
 
-
 def Grayscale(im, desat=(0.2126, 0.7152, 0.0722), **properties):
     """
     :doc: im_im
     :args: (im, **properties)
 
-    An image manipulator that creats a desaturated version of the image
+    An image manipulator that creates a desaturated version of the image
     manipulator `im`.
     """
 
@@ -1416,6 +1614,7 @@ def Alpha(image, alpha, **properties):
 
     return Recolor(image, 255, 255, 255, int(255 * alpha), force_alpha=True, **properties)
 
+
 class Tile(ImageBase):
     """
     :doc: im_im
@@ -1436,8 +1635,8 @@ class Tile(ImageBase):
         self.image = im
         self.size = size
 
-    def get_mtime(self):
-        return self.image.get_mtime()
+    def get_hash(self):
+        return self.image.get_hash()
 
     def load(self):
 
@@ -1462,6 +1661,7 @@ class Tile(ImageBase):
     def predict_files(self):
         return self.image.predict_files()
 
+
 class AlphaMask(ImageBase):
     """
     :doc: im_im
@@ -1474,6 +1674,9 @@ class AlphaMask(ImageBase):
     image, like having one jpeg for color data, and a second one
     for alpha. In some cases, two jpegs can be smaller than a
     single png file.
+
+    Note that this takes different arguments from :func:`AlphaMask`,
+    which uses the mask's alpha channel.
     """
 
     def __init__(self, base, mask, **properties):
@@ -1482,8 +1685,8 @@ class AlphaMask(ImageBase):
         self.base = image(base)
         self.mask = image(mask)
 
-    def get_mtime(self):
-        return max(self.base.get_mtime(), self.image.get_mtime())
+    def get_hash(self):
+        return self.base.get_hash() + self.image.get_hash()
 
     def load(self):
 
@@ -1502,6 +1705,7 @@ class AlphaMask(ImageBase):
     def predict_files(self):
         return self.base.predict_files() + self.mask.predict_files()
 
+
 def image(arg, loose=False, **properties):
     """
     :doc: im_image
@@ -1553,16 +1757,30 @@ def image(arg, loose=False, **properties):
         raise Exception("Could not construct image from argument.")
 
 
-def load_image(fn):
+def load_image(im):
     """
-    This loads an image from the given filename, using the cache.
+    :name: renpy.load_image
+    :doc: udd_utility
+
+    Loads the image manipulator `im` using the image cache, and returns a render.
     """
 
-    surf = cache.get(image(fn))
-    return renpy.display.draw.load_texture(surf)
+    return cache.get(image(im), render=True)
+
+
+def load_surface(im):
+    """
+    :name: renpy.load_surface
+    :doc: udd_utility
+
+    Loads the image manipulator `im` using the image cache, and returns a pygame Surface.
+    """
+
+    return cache.get(image(im))
+
 
 def reset_module():
-    print "Resetting cache."
+    print("Resetting cache.")
 
     global cache
     cache = Cache()
