@@ -1,4 +1,4 @@
-# Copyright 2004-2014 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,12 +19,15 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import print_function
+
 import renpy.display
 from renpy.display.render import render, Render, Matrix2D
 
 # This file contains displayables that are image-like, because they take
 # up a rectangular area of the screen, and do not respond to input.
 
+
 class Solid(renpy.display.core.Displayable):
     """
     :doc: disp_imagelike
@@ -46,11 +49,23 @@ class Solid(renpy.display.core.Displayable):
         else:
             self.color = None
 
+    def __hash__(self):
+        return hash(self.color)
+
+    def __eq__(self, o):
+        if not self._equals(o):
+            return False
+
+        return (self.color == o.color)
+
     def visit(self):
         return [ ]
 
     def render(self, width, height, st, at):
 
+        width = max(self.style.xminimum, width)
+        height = max(self.style.yminimum, height)
+
         color = self.color or self.style.color
 
         rv = Render(width, height)
@@ -71,13 +86,62 @@ class Solid(renpy.display.core.Displayable):
 
         return rv
 
+
+class Borders(object):
+    """
+    :doc: disp_imagelike
+
+    This object provides border size and tiling information to a :func:`Frame`.
+    It can also provide padding information that can be supplied to the
+    :propref:`padding` style property of a window or frame.
+
+    `left`, `top`, `right`, `bottom`
+        These provide the size of the insets used by a frame, and are added
+        to the padding on each side. They should zero or a positive integer.
+
+    `pad_left`, `pad_top`, `pad_right`, `pad_bottom`
+        These are added to the padding on each side, and may be positive or
+        negative. (For example, if `left` is 5 and `pad_left` is -3, the final
+        padding is 2.)
+
+    The padding information is supplied via a field:
+
+    .. attribute:: padding
+
+        This is a four-element tuple containing the padding on each of the
+        four sides.
+    """
+
+    def __init__(self, left, top, right, bottom, pad_left=0, pad_top=0, pad_right=0, pad_bottom=0):
+
+        self.left = left
+        self.top = top
+        self.right = right
+        self.bottom = bottom
+
+        self.pad_left = pad_left
+        self.pad_top = pad_top
+        self.pad_right = pad_right
+        self.pad_bottom = pad_bottom
+
+    @property
+    def padding(self):
+        return (
+            self.left + self.pad_left,
+            self.top + self.pad_top,
+            self.right + self.pad_right,
+            self.bottom + self.pad_bottom,
+            )
+
+
 class Frame(renpy.display.core.Displayable):
     """
     :doc: disp_imagelike
-    :args: (image, left, top, right=None, bottom=None, tile=False, **properties)
+    :args: (image, left=0, top=0, right=None, bottom=None, tile=False, **properties)
+    :name: Frame
 
     A displayable that resizes an image to fill the available area,
-    while preserving the width and height of its borders.  is often
+    while preserving the width and height of its borders. It is often
     used as the background of a window or button.
 
     .. figure:: frame_example.png
@@ -88,7 +152,9 @@ class Frame(renpy.display.core.Displayable):
         An image manipulator that will be resized by this frame.
 
     `left`
-        The size of the border on the left side.
+        The size of the border on the left side. This can also be a
+        :func:`Borders` object, in which case that object is used in place
+        of the other parameters.
 
     `top`
         The size of the border on the top.
@@ -101,8 +167,10 @@ class Frame(renpy.display.core.Displayable):
         The side of the border on the bottom. If None, defaults to `top`.
 
     `tile`
-        If true, tiling is used to resize sections of the image,
-        rather than scaling.
+        If set to True, tiling is used to resize sections of the image,
+        rather than scaling. If set to the string "integer", the nearest
+        integer number of tiles will be used in each direction. That set of
+        full tiles will then be scaled up or down to fit the required area.
 
     ::
 
@@ -113,6 +181,9 @@ class Frame(renpy.display.core.Displayable):
 
     __version__ = 1
 
+    properties = { }
+    tile_ratio = 0.5
+
     def after_upgrade(self, version):
         if version < 2:
             self.left = self.xborder
@@ -120,17 +191,32 @@ class Frame(renpy.display.core.Displayable):
             self.top = self.yborder
             self.bottom = self.yborder
 
-    def __init__(self, image, left=None, top=None, right=None, bottom=None, xborder=None, yborder=None, bilinear=True, tile=False, **properties):
+    def __init__(self, image, left=None, top=None, right=None, bottom=None,
+                 xborder=0, yborder=0, bilinear=True, tile=False,
+                 tile_ratio=0.5, **properties):
         super(Frame, self).__init__(**properties)
 
         self.image = renpy.easy.displayable(image)
+        self._duplicatable = self.image._duplicatable
+
+        if isinstance(left, Borders):
+            insets = left
+
+            left = insets.left
+            top = insets.top
+            right = insets.right
+            bottom = insets.bottom
+
         self.tile = tile
+        # When tile="integer" the proportion of an edge tile that determines
+        # whether to use less tiles and stretch or more tiles and shrink
+        self.tile_ratio = float(tile_ratio)
 
         # Compat for old argument names.
         if left is None:
             left = xborder
         if top is None:
-            top= yborder
+            top = yborder
 
         if right is None:
             right = left
@@ -142,9 +228,48 @@ class Frame(renpy.display.core.Displayable):
         self.right = right
         self.bottom = bottom
 
+    def __repr__(self):
+        return "<Frame {!r} ({},{},{},{}){}>".format(
+            self.image,
+            self.left,
+            self.top,
+            self.right,
+            self.bottom,
+            (" tile ({})".format(self.tile_ratio) if self.tile == "integer"
+             else " tile" if self.tile
+             else ""))
+
+    def __eq__(self, o):
+        if not self._equals(o):
+            return False
+
+        if self.image != o.image:
+            return False
+
+        if self.left != o.left:
+            return False
+        if self.top != o.top:
+            return False
+        if self.right != o.right:
+            return False
+        if self.bottom != o.bottom:
+            return False
+
+        if self.tile != o.tile:
+            return False
+
+        if self.tile_ratio != o.tile_ratio:
+            return False
+
+        return True
+
     def render(self, width, height, st, at):
 
-        crend = render(self.image, width, height, st, at)
+        width = max(self.style.xminimum, width)
+        height = max(self.style.yminimum, height)
+
+        image = self.style.child or self.image
+        crend = render(image, width, height, st, at)
 
         sw, sh = crend.get_size()
         sw = int(sw)
@@ -157,7 +282,7 @@ class Frame(renpy.display.core.Displayable):
         bh = self.top + self.bottom
 
         xborder = min(bw, sw - 2, dw)
-        if xborder:
+        if xborder and bw:
             left = self.left * xborder / bw
             right = self.right * xborder / bw
         else:
@@ -165,7 +290,7 @@ class Frame(renpy.display.core.Displayable):
             right = 0
 
         yborder = min(bh, sh - 2, dh)
-        if yborder:
+        if yborder and bh:
             top = self.top * yborder / bh
             bottom = self.bottom * yborder / bh
         else:
@@ -233,17 +358,36 @@ class Frame(renpy.display.core.Displayable):
             if csw != cdw or csh != cdh:
 
                 if self.tile:
-                    newcr = Render(cdw, cdh)
-                    newcr.clipping = True
+                    ctw, cth = cdw, cdh
 
-                    for x in xrange(0, cdw, csw):
-                        for y in xrange(0, cdh, csh):
-                            newcr.blit(cr, (x, y))
+                    xtiles = max(1, cdw // csw + (1 if cdw % csw else 0))
+                    ytiles = max(1, cdh // csh + (1 if cdh % csh else 0))
 
+                    if cdw % csw or cdh % csh:
+                        # Area is not an exact integer number of tiles
+
+                        if self.tile == "integer":
+                            if cdw % csw / float(csw) < self.tile_ratio:
+                                xtiles = max(1, xtiles-1)
+                            if cdh % csh / float(csh) < self.tile_ratio:
+                                ytiles = max(1, ytiles-1)
+
+                            # Set size of the used tiles (ready to scale)
+                            ctw, cth = csw * xtiles, csh * ytiles
+
+                    newcr = Render(ctw, cth)
+                    newcr.xclipping = True
+                    newcr.yclipping = True
+
+                    for x in xrange(0, xtiles):
+                        for y in xrange(0, ytiles):
+                            newcr.blit(cr, (x * csw, y * csh))
+
+                    csw, csh = ctw, cth
                     cr = newcr
 
-                else:
-
+                if csw != cdw or csh != cdh:
+                    # Subsurface needs scaling
                     newcr = Render(cdw, cdh)
                     newcr.forward = Matrix2D(1.0 * csw / cdw, 0, 0, 1.0 * csh / cdh)
                     newcr.reverse = Matrix2D(1.0 * cdw / csw, 0, 0, 1.0 * cdh / csh)
@@ -292,8 +436,6 @@ class Frame(renpy.display.core.Displayable):
             if right:
                 draw(-right, 0, -bottom, 0)
 
-
-
     def sw_render(self, crend, dw, dh, left, top, right, bottom):
 
         source = crend.render_to_texture(True)
@@ -354,19 +496,42 @@ class Frame(renpy.display.core.Displayable):
 
             # Scale or tile if we have to.
             if dstsize != srcsize:
+
                 if self.tile:
                     tilew, tileh = srcsize
                     dstw, dsth = dstsize
 
-                    surf2 = renpy.display.pgrender.surface_unscaled(dstsize, surf)
+                    xtiles = max(
+                        1, dstw // tilew + (1 if dstw % tilew else 0))
+                    ytiles = max(
+                        1, dsth // tileh + (1 if dsth % tileh else 0))
 
-                    for y in range(0, dsth, tileh):
-                        for x in range(0, dstw, tilew):
-                            surf2.blit(surf, (x, y))
+                    if dstw % tilew or dsth % tileh:
+                        # Area is not an exact integer number of tiles
 
-                    surf = surf2
+                        if self.tile == "integer":
+                            if dstw % tilew / float(tilew) < self.tile_ratio:
+                                xtiles = max(1, xtiles-1)
+                            if dsth % tileh / float(tileh) < self.tile_ratio:
+                                ytiles = max(1, ytiles-1)
 
-                else:
+                    # Tile at least one tile in each direction
+                    surf2 = renpy.display.pgrender.surface_unscaled(
+                        (tilew * xtiles, tileh * ytiles), surf)
+
+                    for y in range(0, ytiles):
+                        for x in range(0, xtiles):
+                            surf2.blit(surf, (x*tilew, y*tileh))
+
+                    if self.tile is True:
+                        # Trim the tiled surface to required size
+                        surf = surf2.subsurface((0, 0, dstw, dsth))
+                    else:
+                        # Using integer full 'tiles' per side
+                        srcsize = (tilew * xtiles, tileh * ytiles)
+                        surf = surf2
+
+                if dstsize != srcsize:
                     surf2 = renpy.display.scale.real_transform_scale(surf, dstsize)
                     surf = surf2
 
@@ -382,8 +547,33 @@ class Frame(renpy.display.core.Displayable):
         # And, finish up.
         return rrv
 
+    def _duplicate(self, args):
+        image = self.image._duplicate(args)
+
+        if image is self.image:
+            return self
+
+        image._unique()
+
+        rv = self._copy(args)
+        rv.image = image
+        rv._duplicatable = image._duplicatable
+        return rv
+
+    def _in_current_store(self):
+        image = self.image._in_current_store()
+
+        if image is self.image:
+            return self
+
+        rv = self._copy()
+        rv.image = image
+        return rv
+
     def visit(self):
-        return [ self.image ]
+        rv = [ ]
+        self.style._visit_frame(rv)
+        return rv
 
 
 class FileCurrentScreenshot(renpy.display.core.Displayable):
@@ -407,7 +597,6 @@ class FileCurrentScreenshot(renpy.display.core.Displayable):
 
         self.empty = empty
 
-
     def render(self, width, height, st, at):
 
         ss = renpy.display.interface.screenshot_surface
@@ -422,4 +611,3 @@ class FileCurrentScreenshot(renpy.display.core.Displayable):
         rv.blit(tex, (0, 0))
 
         return rv
-
