@@ -1,4 +1,4 @@
-# Copyright 2004-2014 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,18 +19,21 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import print_function
+
 import renpy.display
-import pygame
+import pygame_sdl2 as pygame
 import math
 import weakref
 import time
 import os
 
-from renpy.display.render import blit_lock, IDENTITY, BLIT, DISSOLVE, IMAGEDISSOLVE, PIXELLATE
+from renpy.display.render import blit_lock, IDENTITY, BLIT, DISSOLVE, IMAGEDISSOLVE, PIXELLATE, FLATTEN
 
 # A map from cached surface to rle version of cached surface.
 rle_cache = weakref.WeakKeyDictionary()
 
+
 class Clipper(object):
     """
     This is used to calculate the clipping rectangle and update rectangles
@@ -171,7 +174,6 @@ class Clipper(object):
         while sized:
             area, x0, y0, x1, y1 = sized.pop()
 
-
             merged = False
 
             if nca + area >= sa:
@@ -232,8 +234,10 @@ class Clipper(object):
 
         return (x0, y0, x1 - x0, y1 - y0), updates
 
+
 clippers = [ Clipper() ]
 
+
 def surface(w, h, alpha):
     """
     Creates a surface that shares a pixel format with the screen. The created
@@ -247,13 +251,15 @@ def surface(w, h, alpha):
 
     return rv.subsurface((2, 2, w, h))
 
+
 def copy_surface(surf):
     w, h = surf.get_size()
     rv = surface(w, h, True)
 
-    renpy.display.accelerator.nogil_copy(surf, rv) # @UndefinedVariable
+    renpy.display.accelerator.nogil_copy(surf, rv)  # @UndefinedVariable
     return rv
 
+
 def draw_special(what, dest, x, y):
     """
     This handles the special drawing operations, such as dissolve and
@@ -323,7 +329,7 @@ def draw_special(what, dest, x, y):
 
     elif what.operation == PIXELLATE:
 
-        surf = what.children[0][0].render_to_texture(False)
+        surf = what.children[0][0].render_to_texture(dest.get_masks()[3])
 
         px = what.operation_parameter
 
@@ -332,6 +338,10 @@ def draw_special(what, dest, x, y):
             dest.subsurface((0, 0, w, h)),
             px, px, px, px)
 
+    elif what.operation == FLATTEN:
+        surf = what.children[0][0].render_to_texture(dest.get_masks()[3])
+        dest.subsurface((0, 0, w, h)).blit(surf, (0, 0))
+
     else:
         raise Exception("Unknown operation: %d" % what.operation)
 
@@ -378,6 +388,9 @@ def draw(dest, clip, what, xo, yo, screen):
 
         return
 
+    if what.text_input:
+        renpy.display.interface.text_rect = what.screen_rect(xo, yo, None)
+
     # Deal with draw functions.
     if what.operation != BLIT:
 
@@ -425,11 +438,10 @@ def draw(dest, clip, what, xo, yo, screen):
             # what.draw_func(newdest, newx, newy)
             draw_special(what, newdest, newx, newy)
 
-
         return
 
     # Deal with clipping, if necessary.
-    if what.clipping:
+    if what.xclipping or what.yclipping:
 
         if clip:
             cx0, cy0, cx1, cy1 = clip
@@ -487,10 +499,11 @@ def draw(dest, clip, what, xo, yo, screen):
     for child, cxo, cyo, _focus, _main in what.visible_children:
         draw(dest, clip, child, xo + cxo, yo + cyo, screen)
 
+
 def draw_transformed(dest, clip, what, xo, yo, alpha, forward, reverse):
 
     # If our alpha has hit 0, don't do anything.
-    if alpha <= 0.003: # (1 / 256)
+    if alpha <= 0.003:  # (1 / 256)
         return
 
     if forward is None:
@@ -558,7 +571,10 @@ def draw_transformed(dest, clip, what, xo, yo, alpha, forward, reverse):
 
         return
 
-    if what.clipping:
+    if what.text_input:
+        renpy.display.interface.text_rect = what.screen_rect(xo, yo, reverse)
+
+    if what.xclipping or what.yclipping:
 
         if reverse.xdy or reverse.ydx:
             draw_transformed(dest, clip, what.pygame_surface(True), xo, yo, alpha, forward, reverse)
@@ -623,8 +639,8 @@ def draw_transformed(dest, clip, what, xo, yo, alpha, forward, reverse):
         cxo, cyo = reverse.transform(cxo, cyo)
 
         if what.forward:
-            child_forward = forward * what.forward
-            child_reverse = what.reverse * reverse
+            child_forward = what.forward * forward
+            child_reverse = reverse * what.reverse
         else:
             child_forward = forward
             child_reverse = reverse
@@ -632,7 +648,6 @@ def draw_transformed(dest, clip, what, xo, yo, alpha, forward, reverse):
         draw_transformed(dest, clip, child, xo + cxo, yo + cyo, alpha * what.alpha * what.over, child_forward, child_reverse)
 
 
-
 def do_draw_screen(screen_render, full_redraw, swdraw):
     """
     Draws the render produced by render_screen to the screen.
@@ -684,13 +699,9 @@ class SWDraw(object):
         self.mouse_backing_pos = None
         self.mouse_info = None
 
-
         # Is the mouse currently visible?
         self.mouse_old_visible = None
 
-        # This is used to cache the surface->texture operation.
-        self.texture_cache = weakref.WeakKeyDictionary()
-
         # This is used to display video to the screen.
         self.fullscreen_surface = None
 
@@ -716,6 +727,12 @@ class SWDraw(object):
         # 1px border around it iff we're scaling.
         self.window = None
 
+        # Did we show fullscreen video in the last frame?
+        self.showing_video = False
+
+    def get_texture_size(self):
+        return 0, 0
+
     def set_mode(self, virtual_size, physical_size, fullscreen):
 
         # Reset before resize.
@@ -743,23 +760,13 @@ class SWDraw(object):
         else:
             fsflag = 0
 
-        # If a window exists of the right size and flags, use it. Otherwise,
-        # make our own window.
-        old_screen = pygame.display.get_surface()
+        # Don't reuse the old screen, because doing so fails to update
+        # properly on Xorg.
 
         scaled_width = int(width * scale_factor)
         scaled_height = int(height * scale_factor)
 
-        if ((old_screen is not None) and
-            (old_screen.get_size() == (scaled_width, scaled_height)) and
-            (old_screen.get_flags() & pygame.FULLSCREEN == fsflag) and
-            not (old_screen.get_flags() & pygame.OPENGL)
-            ):
-
-            self.screen = old_screen
-
-        else:
-            self.screen = pygame.display.set_mode((scaled_width, scaled_height), fsflag, 32)
+        self.screen = pygame.display.set_mode((scaled_width, scaled_height), fsflag, 32)
 
         if scale_factor != 1.0:
             self.window = surface(width, height, True)
@@ -768,6 +775,11 @@ class SWDraw(object):
 
         renpy.display.pgrender.set_rgba_masks()
 
+        # Scale from the rtt size to the virtual size.
+        self.draw_per_virt = 1.0
+        self.virt_to_draw = renpy.display.render.Matrix2D(self.draw_per_virt, 0, 0, self.draw_per_virt)
+        self.draw_to_virt = renpy.display.render.Matrix2D(1.0 / self.draw_per_virt, 0, 0, 1.0 / self.draw_per_virt)
+
         # Should we redraw the screen from scratch?
         self.full_redraw = True
 
@@ -857,11 +869,21 @@ class SWDraw(object):
         if self.mouse_location:
             updates.append(self.hide_mouse())
 
-        if tex and pos and renpy.game.interface.focused:
+        if tex and pos and renpy.game.interface.mouse_focused:  # @UndefinedVariable
             updates.append(self.show_mouse(pos, info))
 
         return updates
 
+    def translate_point(self, x, y):
+        x /= self.scale_factor
+        y /= self.scale_factor
+        return (x, y)
+
+    def untranslate_point(self, x, y):
+        x *= self.scale_factor
+        y *= self.scale_factor
+        return (x, y)
+
     def update_mouse(self):
         """
         Draws the mouse, and then updates the screen.
@@ -895,7 +917,6 @@ class SWDraw(object):
 
         return pygame.mouse.set_pos([x, y])
 
-
     def screenshot(self, surftree, fullscreen_video):
         """
         Returns a pygame surface containing a screenshot.
@@ -903,7 +924,10 @@ class SWDraw(object):
 
         return self.window
 
-    def should_redraw(self, needs_redraw, first_pass):
+    def can_block(self):
+        return True
+
+    def should_redraw(self, needs_redraw, first_pass, can_block):
         """
         Uses the framerate to determine if we can and should redraw.
         """
@@ -939,47 +963,57 @@ class SWDraw(object):
 
         return True
 
-
     def draw_screen(self, surftree, fullscreen_video):
         """
         Draws the screen.
         """
 
-        if not fullscreen_video:
+        if fullscreen_video:
 
-            updates = [ ]
+            if not self.showing_video:
+                self.window.fill((0, 0, 0, 255))
 
-            updates.extend(self.draw_mouse(False))
+            w, h = self.window.get_size()
+            frame = renpy.display.video.render_movie("movie", w, h)
 
-            damage = do_draw_screen(surftree, self.full_redraw, self)
+            if frame is not None:
+                surftree = frame
 
-            if damage:
-                updates.extend(damage)
-
-            self.full_redraw = False
-
-            if self.window is self.screen:
-
-                updates.extend(self.draw_mouse(True))
-                pygame.display.update(updates)
-
-            else:
-
-                if self.scale_fast:
-                    pygame.transform.scale(self.window, self.screen.get_size(), self.screen)
-                else:
-                    renpy.display.scale.smoothscale(self.window, self.screen.get_size(), self.screen)
-
-                self.draw_mouse(True)
-                pygame.display.flip()
+            self.full_redraw = True
+            self.showing_video = True
 
         else:
+            self.showing_video = False
+
+        updates = [ ]
+
+        updates.extend(self.draw_mouse(False))
+
+        damage = do_draw_screen(surftree, self.full_redraw, self)
+
+        if damage:
+            updates.extend(damage)
+
+        self.full_redraw = False
+
+        if self.window is self.screen:
+
+            updates.extend(self.draw_mouse(True))
+            pygame.display.update(updates)
+
+        else:
+
+            if self.scale_fast:
+                pygame.transform.scale(self.window, self.screen.get_size(), self.screen)
+            else:
+                renpy.display.scale.smoothscale(self.window, self.screen.get_size(), self.screen)
+
+            self.draw_mouse(True)
             pygame.display.flip()
+
+        if fullscreen_video:
             self.full_redraw = True
 
-        self.suppressed_blit = fullscreen_video
-
-
     def render_to_texture(self, render, alpha):
 
         rv = surface(render.width, render.height, alpha)
@@ -989,6 +1023,17 @@ class SWDraw(object):
 
     def is_pixel_opaque(self, what, x, y):
 
+        # Special case ImageDissolve/AlphaMask for speed and correctness
+        # reasons.
+        #
+        # This doesn't work perfectly, but this should be a rare case and
+        # swdraw is going away.
+        if what.operation == IMAGEDISSOLVE:
+            a0 = self.is_pixel_opaque(what.visible_children[0][0], x, y)
+            a2 = self.is_pixel_opaque(what.visible_children[2][0], x, y)
+
+            return a0 * a2
+
         if x < 0 or y < 0 or x >= what.width or y >= what.height:
             return 0
 
@@ -999,7 +1044,6 @@ class SWDraw(object):
             if what.forward:
                 cx, cy = what.forward.transform(cx, cy)
 
-
             if isinstance(child, renpy.display.render.Render):
                 if self.is_pixel_opaque(child, x, y):
                     return True
@@ -1008,18 +1052,18 @@ class SWDraw(object):
                 cx = int(cx)
                 cy = int(cy)
 
+                if cx < 0 or cy < 0:
+                    return False
+
                 cw, ch = child.get_size()
                 if cx >= cw or cy >= ch:
                     return False
 
-
-
                 if not child.get_masks()[3] or child.get_at((cx, cy))[3]:
                     return True
 
         return False
 
-
     def mutated_surface(self, surf):
         """
         Called to indicate that the given surface has changed.
@@ -1031,7 +1075,6 @@ class SWDraw(object):
         if surf in rle_cache:
             del rle_cache[surf]
 
-
     def load_texture(self, surf, transient=False):
         """
         Creates a texture from the surface. In the software implementation,
@@ -1039,23 +1082,21 @@ class SWDraw(object):
         is in the RLE cache.
         """
 
-        surf = copy_surface(surf)
-        self.mutated_surface(surf)
+        if surf in rle_cache:
+            return rle_cache[surf]
 
-        if transient:
-            return surf
+        rle_surf = copy_surface(surf)
 
-        if renpy.game.less_memory:
-            return surf
-
-        if surf not in rle_cache:
-            rle_surf = copy_surface(surf)
+        if not transient:
             rle_surf.set_alpha(255, pygame.RLEACCEL)
-            self.mutated_surface(rle_surf)
 
-            rle_cache[surf] = rle_surf
+        self.mutated_surface(rle_surf)
+        rle_cache[surf] = rle_surf
 
-        return surf
+        return rle_surf
+
+    def ready_one_texture(self):
+        return False
 
     def solid_texture(self, w, h, color):
         """
@@ -1071,24 +1112,14 @@ class SWDraw(object):
         self.mutated_surface(surf)
         return surf
 
-
-    def free_memory(self):
+    def kill_textures(self):
         """
-        Frees up memory.
+        Kills all textures and caches of textures.
         """
 
         rle_cache.clear()
 
-    def deinit(self):
-        """
-        Called when we're restarted.
-        """
-
-        renpy.display.render.free_memory()
-
-        return
-
-    def quit(self): #@ReservedAssignment
+    def quit(self):  # @ReservedAssignment
         """
         Shuts down the drawing system.
         """
