@@ -1,4 +1,4 @@
-# Copyright 2004-2014 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,6 +19,7 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import print_function
 import renpy.display
 import renpy.text
 import codecs
@@ -28,6 +29,11 @@ import sys
 import collections
 import textwrap
 
+import __builtin__
+
+python_builtins = set(dir(__builtin__))
+renpy_builtins = set()
+
 image_prefixes = None
 
 # Things to check in lint.
@@ -46,6 +52,8 @@ image_prefixes = None
 report_node = None
 
 # Reports a message to the user.
+
+
 def report(msg, *args):
     if report_node:
         out = u"%s:%d " % (renpy.parser.unicode_filename(report_node.filename), report_node.linenumber)
@@ -53,21 +61,43 @@ def report(msg, *args):
         out = ""
 
     out += msg % args
-    print
-    print out.encode('utf-8')
+    print()
+    print(out.encode('utf-8'))
+
 
 added = { }
 
 # Reports additional information about a message, the first time it
 # occurs.
-def add(msg):
+
+
+def add(msg, *args):
     if not msg in added:
         added[msg] = True
-        print unicode(msg).encode('utf-8')
+        msg = unicode(msg) % args
+        print(msg.encode('utf-8'))
 
 
-# Trys to evaluate an expression, announcing an error if it fails.
+# Tries to evaluate an expression, announcing an error if it fails.
 def try_eval(where, expr, additional=None):
+    """
+    :doc: lint
+
+    Tries to evaluate an expression, and writes an error to lint.txt if
+    it fails.
+
+    `where`
+        A string giving the location the expression is found. Used to
+        generate an error message of the form "Could not evaluate `expr`
+        in `where`."
+
+    `expr`
+        The expression to try evaluating.
+
+    `additional`
+        If given, an additional line of information that is addded to the
+        error message.
+    """
 
     # Make sure the expression compiles.
     try_compile(where, expr)
@@ -85,13 +115,33 @@ def try_eval(where, expr, additional=None):
     if m.group(1) in __builtins__:
         return
 
-    report( "Could not evaluate '%s', in %s.", expr, where)
+    report("Could not evaluate '%s', in %s.", expr, where)
     if additional:
         add(additional)
 
 # Returns True of the expression can be compiled as python, False
 # otherwise.
+
+
 def try_compile(where, expr, additional=None):
+    """
+    :doc: lint
+
+    Tries to compile an expression, and writes an error to lint.txt if
+    it fails.
+
+    `where`
+        A string giving the location the expression is found. Used to
+        generate an error message of the form "Could not evaluate `expr`
+        in `where`."
+
+    `expr`
+        The expression to try compiling.
+
+    `additional`
+        If given, an additional line of information that is addded to the
+        error message.
+    """
 
     try:
         renpy.python.py_compile_eval_bytecode(expr)
@@ -101,10 +151,17 @@ def try_compile(where, expr, additional=None):
             add(additional)
 
 
-# The sets of names + attributes that
+# The sets of names + attributes that we know are valid.
 imprecise_cache = set()
 
+
 def image_exists_imprecise(name):
+    """
+    Returns true if the image is a plausible image that can be used in a show
+    statement. This returns true if at least one image exists with the same
+    tag and containing all of the attributes (and none of the removed attributes).
+    """
+
     if name in imprecise_cache:
         return True
 
@@ -119,17 +176,22 @@ def image_exists_imprecise(name):
         else:
             required.add(i)
 
-    for im in renpy.display.image.images:
+    for im, d in renpy.display.image.images.items():
 
         if im[0] != nametag:
             continue
 
         attrs = set(im[1:])
 
-        if [ i for i in required if i not in attrs ]:
+        if [ i for i in banned if i in attrs ]:
             continue
 
-        if [ i for i in banned if i in attrs ]:
+        li = getattr(d, "_list_attributes", None)
+
+        if li is not None:
+            attrs = attrs | set(li(im[0], required))
+
+        if [ i for i in required if i not in attrs ]:
             continue
 
         imprecise_cache.add(name)
@@ -138,6 +200,58 @@ def image_exists_imprecise(name):
     return False
 
 
+precise_cache = set()
+
+
+def image_exists_precise(name):
+    """
+    Returns true if an image exists with the same tag and attributes as
+    `name`. (The attributes are allowed to occur in any order.)
+    """
+
+    if name in precise_cache:
+        return True
+
+    nametag = name[0]
+
+    required = set()
+    banned = set()
+
+    for i in name[1:]:
+        if i[0] == "-":
+            banned.add(i[1:])
+        else:
+            required.add(i)
+
+    for im, d in renpy.display.image.images.items():
+
+        if im[0] != nametag:
+            continue
+
+        attrs = set(im[1:])
+
+        if attrs - required:
+            continue
+
+        rest = required - attrs
+
+        if rest:
+
+            try:
+                da = renpy.display.core.DisplayableArguments()
+                da.name=( im[0], ) + tuple(i for i in name[1:] if i in attrs)
+                da.args=tuple(i for i in name[1:] if i in rest)
+                da.lint = True
+                d._duplicate(da)
+            except:
+                continue
+
+        precise_cache.add(name)
+
+        return True
+
+    return False
+
 
 # This reports an error if we're sure that the image with the given name
 # does not exist.
@@ -146,7 +260,6 @@ def image_exists(name, expression, tag, precise=True):
     Checks a scene or show statement for image existence.
     """
 
-
     # Add the tag to the set of known tags.
     tag = tag or name[0]
     image_prefixes[tag] = True
@@ -154,29 +267,22 @@ def image_exists(name, expression, tag, precise=True):
     if expression:
         return
 
-    namelist = list(name)
-    names = " ".join(namelist)
-
-    # Look for the precise name.
-    while namelist:
-        if tuple(namelist) in renpy.display.image.images:
+    if not precise:
+        if image_exists_imprecise(name):
             return
 
-        namelist.pop()
-
     # If we're not precise, then we have to start looking for images
     # that we can possibly match.
-    if not precise and image_exists_imprecise(name):
+    if image_exists_precise(name):
         return
 
-    report("The image named '%s' was not declared.", names)
-
-
+    report("'%s' is not an image.", " ".join(name))
 
 
 # Only check each file once.
 check_file_cache = { }
 
+
 def check_file(what, fn):
 
     present = check_file_cache.get(fn, None)
@@ -203,8 +309,11 @@ def check_displayable(what, d):
 
     files = [ ]
 
-    if isinstance(d, renpy.display.core.Displayable):
-        d.visit_all(lambda a : a.predict_one())
+    try:
+        if isinstance(d, renpy.display.core.Displayable):
+            d.visit_all(lambda a: a.predict_one())
+    except:
+        pass
 
     for fn in files:
         check_file(what, fn)
@@ -217,6 +326,7 @@ def check_image(node):
 
     check_displayable('image %s' % name, renpy.display.image.images[node.imgname])
 
+
 def imspec(t):
     if len(t) == 3:
         return t[0], None, None, t[1], t[2], 0
@@ -226,7 +336,7 @@ def imspec(t):
         return t
 
 
-# Lints ast.Show and ast.Scene nodets.
+# Lints ast.Show and ast.Scene nodes.
 def check_show(node, precise):
 
     # A Scene may have an empty imspec.
@@ -235,13 +345,24 @@ def check_show(node, precise):
 
     name, expression, tag, at_list, layer, _zorder, _behind = imspec(node.imspec)
 
+    layer = renpy.exports.default_layer(layer, tag or name)
+
     if layer not in renpy.config.layers and layer not in renpy.config.top_layers:
         report("Uses layer '%s', which is not in config.layers.", layer)
 
     image_exists(name, expression, tag, precise=precise)
 
     for i in at_list:
-        try_eval("the at list of a scene or show statment", i, "Perhaps you forgot to declare, or misspelled, a position?")
+        try_eval("the at list of a scene or show statment", i, "Perhaps you forgot to define or misspelled a transform.")
+
+
+def precheck_show(node):
+    # A Scene may have an empty imspec.
+    if not node.imspec:
+        return
+
+    tag = imspec(node.imspec)[2]
+    image_prefixes[tag] = True
 
 
 # Lints ast.Hide.
@@ -252,18 +373,19 @@ def check_hide(node):
 
     tag = tag or name[0]
 
+    layer = renpy.exports.default_layer(layer, tag)
+
     if layer not in renpy.config.layers and layer not in renpy.config.top_layers:
         report("Uses layer '%s', which is not in config.layers.", layer)
 
     if tag not in image_prefixes:
         report("The image tag '%s' is not the prefix of a declared image, nor was it used in a show statement before this hide statement.", tag)
 
-    # for i in at_list:
-    #    try_eval(node, "at list of hide statment", i)
 
 def check_with(node):
     try_eval("a with statement or clause", node.expr, "Perhaps you forgot to declare, or misspelled, a transition?")
 
+
 def check_user(node):
 
     def error(msg):
@@ -280,12 +402,13 @@ def check_user(node):
     except:
         report("Didn't properly report what the next statement should be.")
 
+
 def text_checks(s):
     msg = renpy.text.extras.check_text_tags(s)
     if msg:
         report("%s (in %s)", msg, repr(s)[1:])
 
-    if "%" in s:
+    if "%" in s and renpy.config.old_substitutions:
 
         state = 0
         pos = 0
@@ -322,10 +445,16 @@ def text_checks(s):
         if state != 0:
             report("Unterminated string format code '%s' (in %s)", fmt, repr(s)[1:])
 
+
 def check_say(node):
 
     if node.who:
-        try_eval("the who part of a say statement", node.who, "Perhaps you forgot to declare a character?")
+        try:
+            char = renpy.ast.eval_who(node.who)
+        except:
+            report("Could not evaluate '%s' in the who part of a say statement.", node.who)
+            add("Perhaps you forgot to define a character?")
+            char = None
 
     if node.with_:
         try_eval("the with clause of a say statement", node.with_, "Perhaps you forgot to declare, or misspelled, a transition?")
@@ -339,8 +468,6 @@ def check_say(node):
     if node.who is None:
         return
 
-    char = getattr(renpy.store, node.who, None)
-
     if not isinstance(char, renpy.character.ADVCharacter):
         return
 
@@ -352,9 +479,13 @@ def check_say(node):
 
     name = (char.image_tag,) + node.attributes
 
-    if not image_exists_imprecise(name):
-        report("Could not find image (%s) corresponding to attributes on say statement.", " ".join(name))
+    if image_exists_imprecise(name):
+        return
 
+    if image_exists_imprecise(('side', ) + name):
+        return
+
+    report("Could not find image (%s) corresponding to attributes on say statement.", " ".join(name))
 
 
 def check_menu(node):
@@ -371,6 +502,7 @@ def check_menu(node):
 
         text_checks(l)
 
+
 def check_jump(node):
 
     if node.expression:
@@ -379,6 +511,7 @@ def check_jump(node):
     if not renpy.game.script.has_label(node.target):
         report("The jump is to nonexistent label '%s'.", node.target)
 
+
 def check_call(node):
 
     if node.expression:
@@ -387,27 +520,81 @@ def check_call(node):
     if not renpy.game.script.has_label(node.label):
         report("The call is to nonexistent label '%s'.", node.label)
 
+
 def check_while(node):
     try_compile("in the condition of the while statement", node.condition)
 
+
 def check_if(node):
 
     for condition, _block in node.entries:
         try_compile("in a condition of the if statement", condition)
 
+
+def check_define(node, kind):
+    if node.store != 'store':
+        return
+
+    if node.varname in renpy.config.lint_ignore_replaces:
+        return
+
+    if node.varname in python_builtins:
+        report("'%s %s' replaces a python built-in name, which may cause problems.", kind, node.varname)
+
+    if node.varname in renpy_builtins:
+        report("'%s %s' replaces a Ren'Py built-in name, which may cause problems.", kind, node.varname)
+
+
+def check_style_property_displayable(name, property, d):
+
+    if not d._duplicatable:
+        check_displayable(
+            "{}, property {}".format(name, property),
+            d)
+        return
+
+    renpy.style.init_inspect()
+
+    def sort_short(l):
+        l = list(l)
+        l.sort(key=lambda a: len(a))
+        return l
+
+    alts = sort_short(renpy.style.prefix_alts)
+
+    for p in sort_short(renpy.style.affects.get(property, [ ])):
+        for prefix in alts:
+            rest = p[len(prefix):]
+            if rest in renpy.style.all_properties:
+                args = d._args.copy(prefix=prefix)
+                dd = d._duplicate(args)
+                dd._unique()
+
+                check_displayable(
+                    "{}, property {}".format(name, prefix + property),
+                    dd)
+
+                break
+
+        # print property, p
+
+
 def check_style(name, s):
 
     for p in s.properties:
         for k, v in p.iteritems():
 
-            kname = name + ", property " + k
-
             # Treat font specially.
             if k.endswith("font"):
-                check_file(name, v)
+                if isinstance(v, renpy.text.font.FontGroup):
+                    for f in set(v.map.values()):
+                        check_file(name, f)
+                else:
+                    check_file(name, v)
 
             if isinstance(v, renpy.display.core.Displayable):
-                check_displayable(kname, v)
+                check_style_property_displayable(name, k, v)
+
 
 def check_label(node):
 
@@ -428,14 +615,22 @@ def check_label(node):
         add_arg(pi.extrakw)
 
 
+def check_screen(node):
+
+    if (node.screen.parameters is None) and renpy.config.lint_screens_without_parameters:
+        report("The screen %s has not been given a parameter list.", node.screen.name)
+        add("This can be fixed by writing 'screen %s():' instead.", node.screen.name)
+
+
 def check_styles():
-    for name, s in renpy.style.styles.iteritems(): # @UndefinedVariable
-        name = "style." + name[0]
-        for i in name[1:]:
+    for full_name, s in renpy.style.styles.iteritems():  # @UndefinedVariable
+        name = "style." + full_name[0]
+        for i in full_name[1:]:
             name += "[{!r}]".format(i)
 
         check_style("Style " + name, s)
 
+
 def humanize(n):
     s = str(n)
 
@@ -449,6 +644,7 @@ def humanize(n):
 
     return ''.join(rv)
 
+
 def check_filename_encodings():
     """
     Checks files to ensure that they are displayable in unicode.
@@ -464,6 +660,7 @@ def check_filename_encodings():
         report("%s contains non-ASCII characters in its filename.", filename)
         add("(ZIP file distributions can only reliably include ASCII filenames.)")
 
+
 class Count(object):
     """
     Stores information about the word count.
@@ -484,6 +681,7 @@ class Count(object):
         self.words += len(s.split())
         self.characters += len(s)
 
+
 def common(n):
     """
     Returns true if the node is in the common directory.
@@ -496,13 +694,14 @@ def common(n):
     else:
         return False
 
+
 def lint():
     """
     The master lint function, that's responsible for staging all of the
     other checks.
     """
 
-    ap = renpy.arguments.ArgumentParser(description="Checks the script for errors and prints script statistics.")
+    ap = renpy.arguments.ArgumentParser(description="Checks the script for errors and prints script statistics.", require_command=False)
     ap.add_argument("filename", nargs='?', action="store", help="The file to write to.")
 
     args = ap.parse_args()
@@ -513,8 +712,8 @@ def lint():
 
     renpy.game.lint = True
 
-    print codecs.BOM_UTF8
-    print unicode(renpy.version + " lint report, generated at: " + time.ctime()).encode("utf-8")
+    print(codecs.BOM_UTF8)
+    print(unicode(renpy.version + " lint report, generated at: " + time.ctime()).encode("utf-8"))
 
     # This supports check_hide.
     global image_prefixes
@@ -541,6 +740,10 @@ def lint():
 
     global report_node
 
+    for _fn, _ln, node in all_stmts:
+        if isinstance(node, (renpy.ast.Show, renpy.ast.Scene)):
+            precheck_show(node)
+
     for _fn, _ln, node in all_stmts:
 
         if common(node):
@@ -553,10 +756,10 @@ def lint():
             check_image(node)
 
         elif isinstance(node, renpy.ast.Show):
-            check_show(node, True)
+            check_show(node, False)
 
         elif isinstance(node, renpy.ast.Scene):
-            check_show(node, False)
+            check_show(node, True)
 
         elif isinstance(node, renpy.ast.Hide):
             check_hide(node)
@@ -569,7 +772,6 @@ def lint():
 
             counts[language].add(node.what)
 
-
         elif isinstance(node, renpy.ast.Menu):
             check_menu(node)
             menu_count += 1
@@ -600,6 +802,13 @@ def lint():
 
         elif isinstance(node, renpy.ast.Screen):
             screen_count += 1
+            check_screen(node)
+
+        elif isinstance(node, renpy.ast.Define):
+            check_define(node, "define")
+
+        elif isinstance(node, renpy.ast.Default):
+            check_define(node, "default")
 
     report_node = None
 
@@ -609,7 +818,6 @@ def lint():
     for f in renpy.config.lint_hooks:
         f()
 
-
     lines = [ ]
 
     def report_language(language):
@@ -635,11 +843,10 @@ characters per block. """.format(
 
         lines.append(s)
 
-
-    print
-    print
-    print "Statistics:"
-    print
+    print()
+    print()
+    print("Statistics:")
+    print()
 
     languages = list(counts)
     languages.sort()
@@ -651,17 +858,19 @@ characters per block. """.format(
 
     for l in lines:
         for ll in textwrap.wrap(l, 78):
-            print ll.encode("utf-8")
+            print(ll.encode("utf-8"))
 
-        print
+        print()
 
-    print
-    if renpy.config.developer:
-        print "Remember to set config.developer to False before releasing."
-        print
+    for i in renpy.config.lint_stats_callbacks:
+        i()
 
-    print "Lint is not a substitute for thorough testing. Remember to update Ren'Py"
-    print "before releasing. New releases fix bugs and improve compatibility."
+    print()
+    if renpy.config.developer and (renpy.config.original_developer != "auto"):
+        print("Remember to set config.developer to False before releasing.")
+        print()
+
+    print("Lint is not a substitute for thorough testing. Remember to update Ren'Py")
+    print("before releasing. New releases fix bugs and improve compatibility.")
 
     return False
-
