@@ -1,4 +1,4 @@
-# Copyright 2004-2014 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,6 +19,7 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import print_function
 import renpy
 import os.path
 from pickle import loads
@@ -26,19 +27,44 @@ from cStringIO import StringIO
 import sys
 import types
 import threading
+import zlib
+import re
 
 # Ensure the utf-8 codec is loaded, to prevent recursion when we use it
 # to look up filenames.
 u"".encode("utf-8")
 
-################################################################## Asset Loading
+
+# Physical Paths
+
+def get_path(fn):
+    """
+    Returns the path to `fn` relative to the gamedir. If any of the directories
+    leading to `fn` do not exist, tries to create them.
+
+    This always returns a path, but the path may or may not be writable.
+    """
+
+    fn = os.path.join(renpy.config.gamedir, fn)
+    dn = os.path.dirname(fn)
+
+    try:
+        if not os.path.exists(dn):
+            os.makedirs(dn)
+    except:
+        pass
+
+    return fn
+
+# Asset Loading
+
 
 try:
     import android.apk
 
     expansion = os.environ.get("ANDROID_EXPANSION", None)
     if expansion is not None:
-        print "Using expansion file", expansion
+        print("Using expansion file", expansion)
 
         apks = [
             android.apk.APK(apk=expansion, prefix='assets/x-game/'),
@@ -48,7 +74,7 @@ try:
         game_apks = [ apks[0] ]
 
     else:
-        print "Not using expansion file."
+        print("Not using expansion file.")
 
         apks = [
             android.apk.APK(prefix='assets/x-game/'),
@@ -72,17 +98,12 @@ old_config_archives = None
 # A map from lower-case filename to regular-case filename.
 lower_map = { }
 
+
 def index_archives():
     """
     Loads in the indexes for the archive files. Also updates the lower_map.
     """
 
-    # Update lower_map.
-    lower_map.clear()
-
-    for dir, fn in listdirfiles(): #@ReservedAssignment
-        lower_map[fn.lower()] = fn
-
     # Index the archives.
 
     global old_config_archives
@@ -92,6 +113,11 @@ def index_archives():
 
     old_config_archives = renpy.config.archives[:]
 
+    # Update lower_map.
+    lower_map.clear()
+
+    cleardirfiles()
+
     global archives
     archives = [ ]
 
@@ -141,7 +167,11 @@ def index_archives():
         except:
             raise
 
-def walkdir(dir): #@ReservedAssignment
+    for dir, fn in listdirfiles():  # @ReservedAssignment
+        lower_map[fn.lower()] = fn
+
+
+def walkdir(dir):  # @ReservedAssignment
     rv = [ ]
 
     if not os.path.exists(dir) and not renpy.config.developer:
@@ -151,6 +181,11 @@ def walkdir(dir): #@ReservedAssignment
         if i[0] == ".":
             continue
 
+        try:
+            i = renpy.exports.fsdecode(i)
+        except:
+            continue
+
         if os.path.isdir(dir + "/" + i):
             for fn in walkdir(dir + "/" + i):
                 rv.append(i + "/" + fn)
@@ -160,22 +195,56 @@ def walkdir(dir): #@ReservedAssignment
     return rv
 
 
-def listdirfiles(common=True):
+# A list of files that make up the game.
+game_files = [ ]
+
+# A list of files that are in the common directory.
+common_files = [ ]
+
+# A map from filename to if the file is loadable.
+loadable_cache = { }
+
+
+def cleardirfiles():
     """
-    Returns a list of directory, file tuples known to the system. If
-    the file is in an archive, the directory is None.
+    Clears the lists above when the game has changed.
     """
 
-    rv = [ ]
+    global game_files
+    global common_files
+
+    game_files = [ ]
+    common_files = [ ]
+
+
+def scandirfiles():
+    """
+    Scans directories, archives, and apks and fills out game_files and
+    common_files.
+    """
 
     seen = set()
 
-    if common:
-        list_apks = apks
-    else:
-        list_apks = game_apks
+    def add(dn, fn):
+        if fn in seen:
+            return
 
-    for apk in list_apks:
+        if fn.startswith("cache/"):
+            return
+
+        if fn.startswith("saves/"):
+            return
+
+        files.append((dn, fn))
+        seen.add(fn)
+        loadable_cache[fn.lower()] = True
+
+    for apk in apks:
+
+        if apk not in game_apks:
+            files = common_files  # @UnusedVariable
+        else:
+            files = game_files  # @UnusedVariable
 
         for f in apk.list():
 
@@ -183,49 +252,74 @@ def listdirfiles(common=True):
             # to ensure that aapt actually includes every file.
             f = "/".join(i[2:] for i in f.split("/"))
 
-            if f not in seen:
-                rv.append((None, f))
-                seen.add(f)
+            add(None, f)
 
     for i in renpy.config.searchpath:
 
-        if (not common) and (renpy.config.commondir) and (i == renpy.config.commondir):
-            continue
+        if (renpy.config.commondir) and (i == renpy.config.commondir):
+            files = common_files  # @UnusedVariable
+        else:
+            files = game_files  # @UnusedVariable
 
         i = os.path.join(renpy.config.basedir, i)
         for j in walkdir(i):
-            if j not in seen:
-                rv.append((i, j))
-                seen.add(j)
+            add(i, j)
+
+    files = game_files
 
     for _prefix, index in archives:
         for j in index.iterkeys():
-            if j not in seen:
-                rv.append((None, j))
-                seen.add(j)
+            add(None, j)
 
 
-    return rv
+def listdirfiles(common=True):
+    """
+    Returns a list of directory, file tuples known to the system. If
+    the file is in an archive, the directory is None.
+    """
+
+    if (not game_files) and (not common_files):
+        scandirfiles()
+
+    if common:
+        return game_files + common_files
+    else:
+        return list(game_files)
 
 
 class SubFile(object):
 
-    def __init__(self, f, base, length, start):
-        self.f = f
+    def __init__(self, fn, base, length, start):
+        self.fn = fn
+
+        self.f = None
+
         self.base = base
         self.offset = 0
         self.length = length
         self.start = start
 
-        if start is None:
-            self.name = self.f.name
+        if not self.start:
+            self.name = fn
         else:
             self.name = None
 
+    def open(self):
+        self.f = open(self.fn, "rb")
         self.f.seek(self.base)
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, _type, value, tb):
+        self.close()
+        return False
+
     def read(self, length=None):
 
+        if self.f is None:
+            self.open()
+
         maxlength = self.length - self.offset
 
         if length is not None:
@@ -247,6 +341,9 @@ class SubFile(object):
 
     def readline(self, length=None):
 
+        if self.f is None:
+            self.open()
+
         maxlength = self.length - self.offset
         if length is not None:
             length = min(length, maxlength)
@@ -297,7 +394,7 @@ class SubFile(object):
     def __iter__(self):
         return self
 
-    def next(self): #@ReservedAssignment
+    def next(self):  # @ReservedAssignment
         rv = self.readline()
 
         if not rv:
@@ -308,9 +405,11 @@ class SubFile(object):
     def flush(self):
         return
 
-
     def seek(self, offset, whence=0):
 
+        if self.f is None:
+            self.open()
+
         if whence == 0:
             offset = offset
         elif whence == 1:
@@ -333,12 +432,27 @@ class SubFile(object):
         return self.offset
 
     def close(self):
-        self.f.close()
+        if self.f is not None:
+            self.f.close()
+            self.f = None
 
     def write(self, s):
         raise Exception("Write not supported by SubFile")
 
 
+open_file = open
+
+if "RENPY_FORCE_SUBFILE" in os.environ:
+    def open_file(name, mode):
+        f = open(name, mode)
+
+        f.seek(0, 2)
+        length = f.tell()
+        f.seek(0, 0)
+
+        return SubFile(f, 0, length, '')
+
+
 def load_core(name):
     """
     Returns an open python file object of the given type.
@@ -351,6 +465,14 @@ def load_core(name):
         if rv is not None:
             return rv
 
+    # Look for the file directly.
+    if not renpy.config.force_archives:
+        try:
+            fn = transfn(name)
+            return open_file(fn, "rb")
+        except:
+            pass
+
     # Look for the file in the apk.
     for apk in apks:
         prefixed_name = "/".join("x-" + i for i in name.split("/"))
@@ -360,20 +482,12 @@ def load_core(name):
         except IOError:
             pass
 
-    # Look for the file directly.
-    if not renpy.config.force_archives:
-        try:
-            fn = transfn(name)
-            return file(fn, "rb")
-        except:
-            pass
-
     # Look for it in archive files.
     for prefix, index in archives:
         if not name in index:
             continue
 
-        f = file(transfn(prefix + ".rpa"), "rb")
+        afn = transfn(prefix + ".rpa")
 
         data = [ ]
 
@@ -387,10 +501,12 @@ def load_core(name):
             else:
                 offset, dlen, start = t
 
-            rv = SubFile(f, offset, dlen, start)
+            rv = SubFile(afn, offset, dlen, start)
 
         # Compatibility path.
         else:
+            f = file(afn, "rb")
+
             for offset, dlen in index[name]:
                 f.seek(offset)
                 data.append(f.read(dlen))
@@ -402,28 +518,57 @@ def load_core(name):
 
     return None
 
-def get_prefixes():
+
+def check_name(name):
+    """
+    Checks the name to see if it violates any of Ren'Py's rules.
+    """
+
+    if renpy.config.reject_backslash and "\\" in name:
+        raise Exception("Backslash in filename, use '/' instead: %r" % name)
+
+    if renpy.config.reject_relative:
+
+        split = name.split("/")
+
+        if ("." in split) or (".." in split):
+            raise Exception("Filenames may not contain relative directories like '.' and '..': %r" % name)
+
+
+def get_prefixes(tl=True):
     """
     Returns a list of prefixes to search for files.
     """
 
     rv = [ ]
 
-    language = renpy.game.preferences.language
+    if tl:
+        language = renpy.game.preferences.language  # @UndefinedVariable
+    else:
+        language = None
 
-    if language is not None:
-        rv.append(renpy.config.tl_directory + "/" + language + "/")
+    for prefix in renpy.config.search_prefixes:
 
-    rv.append("")
+        if language is not None:
+            rv.append(renpy.config.tl_directory + "/" + language + "/" + prefix)
+
+        rv.append(prefix)
 
     return rv
 
-def load(name):
+
+def load(name, tl=True):
+
+    if renpy.display.predict.predicting:  # @UndefinedVariable
+        if threading.current_thread().name == "MainThread":
+            raise Exception("Refusing to open {} while predicting.".format(name))
 
     if renpy.config.reject_backslash and "\\" in name:
         raise Exception("Backslash in filename, use '/' instead: %r" % name)
 
-    for p in get_prefixes():
+    name = re.sub(r'/+', '/', name).lstrip('/')
+
+    for p in get_prefixes(tl):
         rv = load_core(p + name)
         if rv is not None:
             return rv
@@ -431,8 +576,6 @@ def load(name):
     raise IOError("Couldn't find file '%s'." % name)
 
 
-loadable_cache = { }
-
 def loadable_core(name):
     """
     Returns True if the name is loadable with load, False if it is not.
@@ -443,11 +586,6 @@ def loadable_core(name):
     if name in loadable_cache:
         return loadable_cache[name]
 
-    for apk in apks:
-        prefixed_name = "/".join("x-" + i for i in name.split("/"))
-        if prefixed_name in apk.info:
-            return True
-
     try:
         transfn(name)
         loadable_cache[name] = True
@@ -455,6 +593,12 @@ def loadable_core(name):
     except:
         pass
 
+    for apk in apks:
+        prefixed_name = "/".join("x-" + i for i in name.split("/"))
+        if prefixed_name in apk.info:
+            loadable_cache[name] = True
+            return True
+
     for _prefix, index in archives:
         if name in index:
             loadable_cache[name] = True
@@ -463,8 +607,14 @@ def loadable_core(name):
     loadable_cache[name] = False
     return False
 
+
 def loadable(name):
 
+    name = name.lstrip('/')
+
+    if (renpy.config.loadable_callback is not None) and renpy.config.loadable_callback(name):
+        return True
+
     for p in get_prefixes():
         if loadable_core(p + name):
             return True
@@ -478,11 +628,13 @@ def transfn(name):
     searched directories.
     """
 
-    name = lower_map.get(name.lower(), name)
+    name = name.lstrip('/')
 
     if renpy.config.reject_backslash and "\\" in name:
         raise Exception("Backslash in filename, use '/' instead: %r" % name)
 
+    name = lower_map.get(name.lower(), name)
+
     if isinstance(name, str):
         name = name.decode("utf-8")
 
@@ -497,22 +649,41 @@ def transfn(name):
     raise Exception("Couldn't find file '%s'." % name)
 
 
-def get_mtime(name):
+hash_cache = dict()
+
+
+def get_hash(name):
     """
     Returns the time the file m was last modified, or 0 if it
     doesn't exist or is archived.
     """
 
-    for p in get_prefixes():
-        try:
-            fn = transfn(p + name)
-            return os.path.getmtime(fn)
-        except:
-            pass
+    rv = hash_cache.get(name, None)
+    if rv is not None:
+        return rv
 
-    return 0
+    rv = 0
 
-################################################################# Module Loading
+    try:
+        f = load(name)
+
+        while True:
+            data = f.read(1024 * 1024)
+
+            if not data:
+                break
+
+            rv = zlib.adler32(data, rv)
+
+    except:
+        pass
+
+    hash_cache[name] = rv
+
+    return rv
+
+
+# Module Loading
 
 class RenpyImporter(object):
     """
@@ -521,9 +692,12 @@ class RenpyImporter(object):
     """
 
     def __init__(self, prefix=""):
-        self.prefix = ""
+        self.prefix = prefix
 
-    def translate(self, fullname, prefix=""):
+    def translate(self, fullname, prefix=None):
+
+        if prefix is None:
+            prefix = self.prefix
 
         try:
             fn = (prefix + fullname.replace(".", "/")).decode("utf8")
@@ -560,28 +734,64 @@ class RenpyImporter(object):
         if filename.endswith("__init__.py"):
             mod.__path__ = [ filename[:-len("__init__.py")] ]
 
-        source = load(filename).read().decode("utf8")
-        if source and source[0] == u'\ufeff':
-            source = source[1:]
-        source = source.encode("raw_unicode_escape")
+        for encoding in [ "utf-8", "latin-1" ]:
+
+            try:
+
+                source = load(filename).read().decode(encoding)
+                if source and source[0] == u'\ufeff':
+                    source = source[1:]
+                source = source.encode("raw_unicode_escape")
+
+                source = source.replace("\r", "")
+
+                code = compile(source, filename, 'exec', renpy.python.old_compile_flags, 1)
+                break
+            except:
+                if encoding == "latin-1":
+                    raise
 
-        source = source.replace("\r", "")
-        code = compile(source, filename, 'exec')
         exec code in mod.__dict__
-        return mod
+
+        return sys.modules[fullname]
 
     def get_data(self, filename):
         return load(filename).read()
 
+
+meta_backup = [ ]
+
+
+def add_python_directory(path):
+    """
+    :doc: other
+
+    Adds `path` to the list of paths searched for Python modules and packages.
+    The path should be a string relative to the game directory. This must be
+    called before an import statement.
+    """
+
+    if path and not path.endswith("/"):
+        path = path + "/"
+
+    sys.meta_path.insert(0, RenpyImporter(path))
+
+
 def init_importer():
-    sys.meta_path.append(RenpyImporter())
+    meta_backup[:] = sys.meta_path
+
+    add_python_directory("python-packages/")
+    add_python_directory("")
+
 
 def quit_importer():
-    sys.meta_path.pop()
+    sys.meta_path[:] = meta_backup
 
-#################################################################### Auto-Reload
 
-# This is set to True if autoreload hads detected an autoreload is needed.
+# Auto-Reload
+
+
+# This is set to True if autoreload has detected an autoreload is needed.
 needs_autoreload = False
 
 # A map from filename to mtime, or None if the file doesn't exist.
@@ -599,6 +809,7 @@ auto_lock = threading.Condition()
 # Used to indicate that this file is blacklisted.
 auto_blacklisted = renpy.object.Sentinel("auto_blacklisted")
 
+
 def auto_mtime(fn):
     """
     Gets the mtime of fn, or None if the file does not exist.
@@ -609,16 +820,19 @@ def auto_mtime(fn):
     except:
         return None
 
-def add_auto(fn):
+
+def add_auto(fn, force=False):
     """
     Adds fn as a file we watch for changes. If it's mtime changes or the file
     starts/stops existing, we trigger a reload.
     """
 
+    fn = fn.replace("\\", "/")
+
     if not renpy.autoreload:
         return
 
-    if fn in auto_mtimes:
+    if (fn in auto_mtimes) and (not force):
         return
 
     for e in renpy.config.autoreload_blacklist:
@@ -632,6 +846,7 @@ def add_auto(fn):
     with auto_lock:
         auto_mtimes[fn] = mtime
 
+
 def auto_thread_function():
     """
     This thread sets need_autoreload when necessary.
@@ -656,7 +871,11 @@ def auto_thread_function():
                 continue
 
             if auto_mtime(fn) != mtime:
-                needs_autoreload = True
+
+                with auto_lock:
+                    if auto_mtime(fn) != auto_mtimes[fn]:
+                        needs_autoreload = True
+
 
 def auto_init():
     """
@@ -678,6 +897,7 @@ def auto_init():
     auto_thread.daemon = True
     auto_thread.start()
 
+
 def auto_quit():
     """
     Terminates the autoreload thread.
