@@ -1,4 +1,4 @@
-# Copyright 2004-2014 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -21,6 +21,8 @@
 
 # This file contains functions that load and save the game state.
 
+from __future__ import print_function
+
 import pickle
 import cPickle
 
@@ -32,28 +34,42 @@ import threading
 import types
 import shutil
 import os
+import sys
 
 import renpy
+from renpy import six
 
 from json import dumps as json_dumps
 
 # Dump that chooses which pickle to use:
+
+
 def dump(o, f):
     if renpy.config.use_cpickle:
         cPickle.dump(o, f, cPickle.HIGHEST_PROTOCOL)
     else:
         pickle.dump(o, f, pickle.HIGHEST_PROTOCOL)
 
+
+def dumps(o):
+    if renpy.config.use_cpickle:
+        return cPickle.dumps(o, cPickle.HIGHEST_PROTOCOL)
+    else:
+        return pickle.dumps(o, pickle.HIGHEST_PROTOCOL)
+
+
 def loads(s):
     if renpy.config.use_cpickle:
         return cPickle.loads(s)
     else:
         return pickle.loads(s)
 
+
 # This is used as a quick and dirty way of versioning savegame
 # files.
 savegame_suffix = renpy.savegame_suffix
 
+
 def save_dump(roots, log):
     """
     Dumps information about the save to save_dump.txt. We dump the size
@@ -63,7 +79,6 @@ def save_dump(roots, log):
 
     o_repr_cache = { }
 
-
     def visit(o, path):
         ido = id(o)
 
@@ -95,7 +110,6 @@ def save_dump(roots, log):
         else:
             o_repr = "BAD TYPE <{0}>".format(type(o).__name__)
 
-
         o_repr_cache[ido] = o_repr
 
         if isinstance(o, (int, float, types.NoneType, types.ModuleType, types.ClassType)):
@@ -161,27 +175,128 @@ def save_dump(roots, log):
                 size += 2
                 size += visit(v, "{0}[{1!r}]".format(path, k))
 
-
         f.write("{0: 7d} {1} = {2}\n".format(size, path, o_repr_cache[ido]))
 
         return size
 
-    f = file("save_dump.txt", "w")
+    f, _ = renpy.error.open_error_file("save_dump.txt", "w")
 
     visit(roots, "roots")
     visit(log, "log")
 
     f.close()
 
+
+def find_bad_reduction(roots, log):
+    """
+    Finds objects that can't be reduced properly.
+    """
+
+    seen = set()
+
+    def visit(o, path):
+        ido = id(o)
+
+        if ido in seen:
+            return
+
+        seen.add(ido)
+
+        if isinstance(o, (int, float, types.NoneType, types.ClassType)):
+            return
+
+        if isinstance(o, (tuple, list)):
+            for i, oo in enumerate(o):
+                rv = visit(oo, "{0}[{1!r}]".format(path, i))
+                if rv is not None:
+                    return rv
+
+        elif isinstance(o, dict):
+            for k, v in o.iteritems():
+                rv = visit(v, "{0}[{1!r}]".format(path, k))
+                if rv is not None:
+                    return rv
+
+        elif isinstance(o, types.MethodType):
+            return visit(o.im_self, path + ".im_self")
+
+        elif isinstance(o, types.ModuleType):
+
+            return "{} = {}".format(path, repr(o)[:160])
+
+        else:
+
+            try:
+                reduction = o.__reduce_ex__(2)
+            except:
+
+                import copy
+
+                try:
+                    copy.copy(o)
+                    return None
+                except:
+                    pass
+
+                return "{} = {}".format(path, repr(o)[:160])
+
+            # Gets an element from the reduction, or o if we don't have
+            # such an element.
+            def get(idx, default):
+                if idx < len(reduction) and reduction[idx] is not None:
+                    return reduction[idx]
+                else:
+                    return default
+
+            state = get(2, { })
+            if isinstance(state, dict):
+                for k, v in state.iteritems():
+                    rv = visit(v, path + "." + k)
+                    if rv is not None:
+                        return rv
+            else:
+                rv = visit(state, path + ".__getstate__()")
+                if rv is not None:
+                    return rv
+
+            for i, oo in enumerate(get(3, [])):
+                rv = visit(oo, "{0}[{1}]".format(path, i))
+                if rv is not None:
+                    return rv
+
+            for i in get(4, []):
+
+                if len(i) != 2:
+                    continue
+
+                k, v = i
+
+                rv = visit(v, "{0}[{1!r}]".format(path, k))
+                if rv is not None:
+                    return rv
+
+        return None
+
+    for k, v in roots.items():
+        rv = visit(v, k)
+        if rv is not None:
+            return rv
+
+    return visit(log, "renpy.game.log")
+
+
 ################################################################################
 # Saving
 ################################################################################
 
 # Used to indicate an aborted save, due to the game being mutated
 # while the save is in progress.
+
+
 class SaveAbort(Exception):
     pass
 
+
 def safe_rename(old, new):
     """
     Safely rename old to new.
@@ -190,7 +305,22 @@ def safe_rename(old, new):
     if os.path.exists(new):
         os.unlink(new)
 
-    os.rename(old, new)
+    try:
+        os.rename(old, new)
+    except:
+
+        # If the rename failed, try again.
+        try:
+            os.unlink(new)
+            os.rename(old, new)
+        except:
+
+            # If it fails a second time, give up.
+            try:
+                os.unlink(old)
+            except:
+                pass
+
 
 class SaveRecord(object):
     """
@@ -216,7 +346,7 @@ class SaveRecord(object):
 
         # For speed, copy the file after we've written it at least once.
         if self.first_filename is not None:
-            shutil.copy(self.first_filename, filename_new)
+            shutil.copyfile(self.first_filename, filename_new)
             safe_rename(filename_new, filename)
             return
 
@@ -244,7 +374,6 @@ class SaveRecord(object):
         self.first_filename = filename
 
 
-
 def save(slotname, extra_info='', mutate_flag=False):
     """
     :doc: loadsave
@@ -268,18 +397,36 @@ def save(slotname, extra_info='', mutate_flag=False):
 
     roots = renpy.game.log.freeze(None)
 
+    if renpy.config.save_dump:
+        save_dump(roots, renpy.game.log)
+
     logf = StringIO()
-    dump((roots, renpy.game.log), logf)
+    try:
+        dump((roots, renpy.game.log), logf)
+    except:
+
+        t, e, tb = sys.exc_info()
+
+        if mutate_flag:
+            six.reraise(t, e, tb)
+
+        try:
+            bad = find_bad_reduction(roots, renpy.game.log)
+        except:
+            six.reraise(t, e, tb)
+
+        if bad is None:
+            six.reraise(t, e, tb)
+
+        e.args = ( e.args[0] + ' (perhaps {})'.format(bad), ) + e.args[1:]
+        six.reraise(t, e, tb)
 
     if mutate_flag and renpy.python.mutate_flag:
         raise SaveAbort()
 
-    if renpy.config.save_dump:
-        save_dump(roots, renpy.game.log)
-
     screenshot = renpy.game.interface.get_screenshot()
 
-    json = { "_save_name" : extra_info }
+    json = { "_save_name" : extra_info, "_renpy_version" : list(renpy.version_tuple), "_version" : renpy.config.version }
 
     for i in renpy.config.save_json_callbacks:
         i(json)
@@ -292,7 +439,8 @@ def save(slotname, extra_info='', mutate_flag=False):
     location.scan()
     clear_slot(slotname)
 
-
+# The thread used for autosave.
+autosave_thread = None
 
 # Flag that lets us know if an autosave is in progress.
 autosave_not_running = threading.Event()
@@ -301,7 +449,8 @@ autosave_not_running.set()
 # The number of times autosave has been called without a save occuring.
 autosave_counter = 0
 
-def autosave_thread(take_screenshot):
+
+def autosave_thread_function(take_screenshot):
 
     global autosave_counter
 
@@ -327,7 +476,9 @@ def autosave_thread(take_screenshot):
 
     finally:
         autosave_not_running.set()
-
+        if renpy.emscripten:
+            import emscripten
+            emscripten.syncfs()
 
 
 def autosave():
@@ -358,12 +509,34 @@ def autosave():
 
 
 # This assumes a screenshot has already been taken.
-def force_autosave(take_screenshot=False):
+def force_autosave(take_screenshot=False, block=False):
+    """
+    :doc: other
+
+    Forces a background autosave to occur.
+
+    `take_screenshot`
+        If True, a new screenshot will be taken. If False, the existing
+        screenshot will be used.
+
+    `block`
+        If True, blocks until the autosave completes.
+    """
+
+    global autosave_thread
+
+    if renpy.game.after_rollback or renpy.exports.in_rollback():
+        return
 
     # That is, autosave is running.
     if not autosave_not_running.isSet():
         return
 
+    # Join the autosave thread to clear resources.
+    if autosave_thread is not None:
+        autosave_thread.join()
+        autosave_thread = None
+
     # Do not save if we're in the main menu.
     if renpy.store.main_menu:
         return
@@ -372,10 +545,31 @@ def force_autosave(take_screenshot=False):
     if renpy.store._in_replay:
         return
 
+    if block:
+
+        if renpy.config.auto_save_extra_info:
+            extra_info = renpy.config.auto_save_extra_info()
+        else:
+            extra_info = ""
+
+        cycle_saves("auto-", renpy.config.autosave_slots)
+
+        if take_screenshot:
+            renpy.exports.take_screenshot()
+
+        save("auto-1", extra_info=extra_info)
+
+        return
+
     autosave_not_running.clear()
-    t = threading.Thread(target=autosave_thread, args=(take_screenshot,))
-    t.daemon = True
-    t.start()
+
+    if not renpy.emscripten:
+        autosave_thread = threading.Thread(target=autosave_thread_function, args=(take_screenshot,))
+        autosave_thread.daemon = True
+        autosave_thread.start()
+    else:
+        import emscripten
+        emscripten.async_call(autosave_thread_function, take_screenshot, -1)
 
 
 ################################################################################
@@ -443,11 +637,17 @@ def list_saved_games(regexp=r'.', fast=False):
 
         c = get_cache(s)
 
-        extra_info = c.get_json().get("_save_name", "")
-        screenshot = c.get_screenshot()
-        mtime = c.get_mtime()
+        if c is not None:
+            json = c.get_json()
+            if json is not None:
+                extra_info = json.get("_save_name", "")
+            else:
+                extra_info = ""
 
-        rv.append((s, extra_info, screenshot, mtime))
+            screenshot = c.get_screenshot()
+            mtime = c.get_mtime()
+
+            rv.append((s, extra_info, screenshot, mtime))
 
     return rv
 
@@ -471,9 +671,11 @@ def list_slots(regexp=None):
 
     return slots
 
+
 # A cache for newest slot info.
 newest_slot_cache = { }
 
+
 def newest_slot(regexp=None):
     """
     :doc: loadsave
@@ -518,6 +720,7 @@ def slot_mtime(slotname):
 
     return get_cache(slotname).get_mtime()
 
+
 def slot_json(slotname):
     """
     :doc: loadsave
@@ -528,6 +731,7 @@ def slot_json(slotname):
 
     return get_cache(slotname).get_json()
 
+
 def slot_screenshot(slotname):
     """
     :doc: loadsave
@@ -538,6 +742,7 @@ def slot_screenshot(slotname):
 
     return get_cache(slotname).get_screenshot()
 
+
 def can_load(filename, test=False):
     """
     :doc: loadsave
@@ -552,6 +757,7 @@ def can_load(filename, test=False):
     else:
         return False
 
+
 def load(filename):
     """
     :doc: loadsave
@@ -563,6 +769,7 @@ def load(filename):
     roots, log = loads(location.load(filename))
     log.unfreeze(roots, label="_after_load")
 
+
 def unlink_save(filename):
     """
     :doc: loadsave
@@ -588,6 +795,18 @@ def rename_save(old, new):
     clear_slot(new)
 
 
+def copy_save(old, new):
+    """
+    :doc: loadsave
+
+    Copies the save at `old` to `new`. (Does nothing if `old` does not
+    exist.)
+    """
+
+    location.copy(old, new)
+    clear_slot(new)
+
+
 def cycle_saves(name, count):
     """
     :doc: loadsave
@@ -606,9 +825,11 @@ def cycle_saves(name, count):
 # Cache
 ################################################################################
 
+
 # None is a possible value for some of the attributes.
 unknown = renpy.object.Sentinel("unknown")
 
+
 class Cache(object):
     """
     This represents cached information about a save slot.
@@ -655,10 +876,22 @@ class Cache(object):
 
         return self.screenshot
 
+    def preload(self):
+        """
+        Preloads all the save data (that won't take up a ton of memory).
+        """
+
+        self.get_mtime()
+        self.get_json()
+        self.get_screenshot()
+
+
+
 # A map from slotname to cache object. This is used to cache savegame scan
 # data until the slot changes.
 cache = { }
 
+
 def get_cache(slotname):
 
     rv = cache.get(slotname, None)
@@ -680,6 +913,7 @@ def clear_slot(slotname):
 
     renpy.exports.restart_interaction()
 
+
 def clear_cache():
     """
     Clears the entire cache.
@@ -693,12 +927,19 @@ def clear_cache():
     renpy.exports.restart_interaction()
 
 
+def init():
+    """
+    Scans all the metadata from the save slot cache.
+    """
+
+    for i in list_slots():
+        if not i.startswith("_"):
+            get_cache(i).preload()
+
+
 # Save locations are places where saves are saved to or loaded from, or a
 # collection of such locations. This is the default save location.
 location = None
 
 if False:
     location = renpy.savelocation.FileLocation("blah")
-
-
-
