@@ -1,4 +1,4 @@
-# Copyright 2004-2014 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,6 +19,8 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import print_function
+
 import renpy.display
 import contextlib
 
@@ -31,6 +33,8 @@ filename = None
 new_variable_serial = 0
 
 # Returns the name of a new variable.
+
+
 @contextlib.contextmanager
 def new_variable():
     global new_variable_serial
@@ -39,11 +43,13 @@ def new_variable():
     yield "_%d" % new_variable_serial
     new_variable_serial -= 1
 
+
 def increment_lineno(node, amount):
     for node in ast.walk(node):
         if hasattr(node, 'lineno'):
             node.lineno += amount
 
+
 class LineNumberNormalizer(ast.NodeVisitor):
 
     def __init__(self):
@@ -64,6 +70,7 @@ class LineNumberNormalizer(ast.NodeVisitor):
 # The parser that things are being added to.
 parser = None
 
+
 class Positional(object):
     """
     This represents a positional parameter to a function.
@@ -75,8 +82,6 @@ class Positional(object):
         if parser:
             parser.add(self)
 
-# Used to generate the documentation
-all_keyword_names = set()
 
 class Keyword(object):
     """
@@ -86,11 +91,10 @@ class Keyword(object):
     def __init__(self, name):
         self.name = name
 
-        all_keyword_names.add(self.name)
-
         if parser:
             parser.add(self)
 
+
 STYLE_PREFIXES = [
     '',
     'insensitive_',
@@ -104,6 +108,7 @@ STYLE_PREFIXES = [
     'selected_activate_',
 ]
 
+
 class Style(object):
     """
     This represents a style parameter to a function.
@@ -112,9 +117,6 @@ class Style(object):
     def __init__(self, name):
         self.name = name
 
-        for j in STYLE_PREFIXES:
-            all_keyword_names.add(j + self.name)
-
         if parser:
             parser.add(self)
 
@@ -128,9 +130,6 @@ class PrefixStyle(object):
         self.prefix = prefix
         self.name = name
 
-        for j in STYLE_PREFIXES:
-            all_keyword_names.add(prefix + j + self.name)
-
         if parser:
             parser.add(self)
 
@@ -222,7 +221,6 @@ class Parser(object):
                 rv.extend(c)
                 count += 1
 
-
         return rv
 
     def parse_eval(self, expr, lineno=1):
@@ -236,7 +234,7 @@ class Parser(object):
 
         try:
             rv = ast.parse(expr, 'eval').body[0].value
-        except SyntaxError, e:
+        except SyntaxError as e:
             raise renpy.parser.ParseError(
                 filename,
                 lineno + e[1][1] - 1,
@@ -259,7 +257,7 @@ class Parser(object):
 
         try:
             rv = ast.parse(code, 'exec')
-        except SyntaxError, e:
+        except SyntaxError as e:
 
             raise renpy.parser.ParseError(
                 filename,
@@ -278,6 +276,12 @@ class Parser(object):
 
         return self.parse_eval(expr, lineno)
 
+    def parse_comma_expression(self, l):
+        lineno = l.number
+        expr = l.require(l.comma_expression)
+
+        return self.parse_eval(expr, lineno)
+
     def parse(self, l, name):
         """
         This is expected to parse a function statement, and to return
@@ -295,6 +299,7 @@ class Parser(object):
 # A singleton value.
 many = renpy.object.Sentinel("many")
 
+
 class FunctionStatementParser(Parser):
     """
     This is responsible for parsing function statements.
@@ -366,7 +371,7 @@ class FunctionStatementParser(Parser):
 
             seen_keywords.add(name)
 
-            expr = self.parse_simple_expression(l)
+            expr = self.parse_comma_expression(l)
 
             call_node.keywords.append(
                 ast.keyword(arg=str(name), value=expr),
@@ -488,135 +493,140 @@ all_statements = [ ]
 childbearing_statements = [ ]
 
 position_property_names = [
-        "anchor",
-        "xanchor",
-        "yanchor",
-        "pos",
-        "xpos",
-        "ypos",
-        "align",
-        "xalign",
-        "yalign",
-        "xoffset",
-        "yoffset",
-        "maximum",
-        "xmaximum",
-        "ymaximum",
-        "area",
-        "clipping",
-        "xfill",
-        "yfill",
-        # no center, since it can conflict with the center transform.
-        "xcenter",
-        "ycenter",
-        "xsize",
-        "ysize",
-        "xysize",
-        ]
+    "anchor",
+    "xanchor",
+    "yanchor",
+    "pos",
+    "xpos",
+    "ypos",
+    "align",
+    "xalign",
+    "yalign",
+    "xoffset",
+    "yoffset",
+    "maximum",
+    "xmaximum",
+    "ymaximum",
+    "area",
+    "clipping",
+    "xfill",
+    "yfill",
+    # no center, since it can conflict with the center transform.
+    "xcenter",
+    "ycenter",
+    "xsize",
+    "ysize",
+    "xysize",
+    "alt",
+    "debug",
+    ]
 
 position_properties = [ Style(i) for i in position_property_names ]
 text_position_properties = [ PrefixStyle("text_", i) for i in position_property_names ]
 side_position_properties = [ PrefixStyle("side_", i) for i in position_property_names ]
 
 text_property_names = [
-        "antialias",
-        "vertical",
-        "black_color",
-        "bold",
-        "color",
-        "drop_shadow",
-        "drop_shadow_color",
-        "first_indent",
-        "font",
-        "size",
-        "hyperlink_functions",
-        "italic",
-        "justify",
-        "kerning",
-        "language",
-        "layout",
-        "line_leading",
-        "line_spacing",
-        "minwidth",
-        "min_width",
-        "newline_indent",
-        "outlines",
-        "rest_indent",
-        "ruby_style",
-        "slow_cps",
-        "slow_cps_multiplier",
-        "slow_abortable",
-        "strikethrough",
-        "text_align",
-        "text_y_fudge",
-        "underline",
-        "minimum",
-        "xminimum",
-        "yminimum",
-        ]
+    "antialias",
+    "vertical",
+    "black_color",
+    "bold",
+    "color",
+    "drop_shadow",
+    "drop_shadow_color",
+    "first_indent",
+    "font",
+    "size",
+    "hyperlink_functions",
+    "italic",
+    "justify",
+    "kerning",
+    "language",
+    "layout",
+    "line_leading",
+    "line_spacing",
+    "minwidth",
+    "min_width",
+    "newline_indent",
+    "outlines",
+    "rest_indent",
+    "ruby_style",
+    "slow_cps",
+    "slow_cps_multiplier",
+    "slow_abortable",
+    "strikethrough",
+    "text_align",
+    "text_y_fudge",
+    "underline",
+    "minimum",
+    "xminimum",
+    "yminimum",
+    ]
 
 text_properties = [ Style(i) for i in text_property_names ]
 text_text_properties = [ PrefixStyle("text_", i) for i in text_property_names ]
 
 window_properties = [ Style(i) for i in [
-        "background",
-        "foreground",
-        "left_margin",
-        "right_margin",
-        "bottom_margin",
-        "top_margin",
-        "xmargin",
-        "ymargin",
-        "left_padding",
-        "right_padding",
-        "top_padding",
-        "bottom_padding",
-        "xpadding",
-        "ypadding",
-        "size_group",
-        "minimum",
-        "xminimum",
-        "yminimum",
-        ] ]
+    "background",
+    "foreground",
+    "left_margin",
+    "right_margin",
+    "bottom_margin",
+    "top_margin",
+    "xmargin",
+    "ymargin",
+    "left_padding",
+    "right_padding",
+    "top_padding",
+    "bottom_padding",
+    "xpadding",
+    "ypadding",
+    "size_group",
+    "minimum",
+    "xminimum",
+    "yminimum",
+    ] ]
 
 button_properties = [ Style(i) for i in [
-        "sound",
-        "mouse",
-        "focus_mask",
-        "child",
-        ] ]
+    "sound",
+    "mouse",
+    "focus_mask",
+    "child",
+    "keyboard_focus",
+    ] ]
 
 bar_properties = [ Style(i) for i in [
-        "bar_vertical",
-        "bar_invert",
-        "bar_resizing",
-        "left_gutter",
-        "right_gutter",
-        "top_gutter",
-        "bottom_gutter",
-        "left_bar",
-        "right_bar",
-        "top_bar",
-        "bottom_bar",
-        "thumb",
-        "thumb_shadow",
-        "thumb_offset",
-        "mouse",
-        "unscrollable",
-        ] ]
+    "bar_vertical",
+    "bar_invert",
+    "bar_resizing",
+    "left_gutter",
+    "right_gutter",
+    "top_gutter",
+    "bottom_gutter",
+    "left_bar",
+    "right_bar",
+    "top_bar",
+    "bottom_bar",
+    "thumb",
+    "thumb_shadow",
+    "thumb_offset",
+    "mouse",
+    "unscrollable",
+    "keyboard_focus",
+    ] ]
 
 box_properties = [ Style(i) for i in [
-        "box_layout",
-        "box_wrap",
-        "box_reverse",
-        "order_reverse",
-        "spacing",
-        "first_spacing",
-        "fit_first",
-        "minimum",
-        "xminimum",
-        "yminimum",
-        ] ]
+    "box_layout",
+    "box_wrap",
+    "box_wrap_spacing",
+    "box_reverse",
+    "order_reverse",
+    "spacing",
+    "first_spacing",
+    "fit_first",
+    "minimum",
+    "xminimum",
+    "yminimum",
+    ] ]
 
 ui_properties = [
     Keyword("at"),
@@ -706,6 +716,7 @@ Keyword("default")
 Keyword("length")
 Keyword("allow")
 Keyword("exclude")
+Keyword("copypaste")
 Keyword("prefix")
 Keyword("suffix")
 Keyword("changed")
@@ -725,6 +736,8 @@ Keyword("clicked")
 Keyword("hovered")
 Keyword("unhovered")
 Keyword("alternate")
+Keyword("selected")
+Keyword("sensitive")
 add(ui_properties)
 add(position_properties)
 add(window_properties)
@@ -744,6 +757,8 @@ Keyword("hovered")
 Keyword("unhovered")
 Keyword("alternate")
 Keyword("image_style")
+Keyword("selected")
+Keyword("sensitive")
 add(ui_properties)
 add(position_properties)
 add(window_properties)
@@ -759,6 +774,8 @@ Keyword("alternate")
 Keyword("text_style")
 Keyword("substitute")
 Keyword("scope")
+Keyword("selected")
+Keyword("sensitive")
 add(ui_properties)
 add(position_properties)
 add(window_properties)
@@ -792,6 +809,7 @@ for name in [ "bar", "vbar" ]:
 FunctionStatementParser("viewport", "ui.viewport", 1)
 Keyword("child_size")
 Keyword("mousewheel")
+Keyword("arrowkeys")
 Keyword("draggable")
 Keyword("edgescroll")
 Keyword("xadjustment")
@@ -865,22 +883,26 @@ Keyword("droppable")
 Keyword("drag_raise")
 Keyword("dragged")
 Keyword("dropped")
+Keyword("drop_allowable")
 Keyword("drag_handle")
 Keyword("drag_joined")
 Keyword("clicked")
 Keyword("hovered")
 Keyword("unhovered")
+Keyword("mouse_drop")
 Style("child")
 add(ui_properties)
 add(position_properties)
 
 FunctionStatementParser("draggroup", "ui.draggroup", many)
+Keyword("min_overlap")
 add(ui_properties)
 add(position_properties)
 
 FunctionStatementParser("mousearea", "ui.mousearea", 0)
 Keyword("hovered")
 Keyword("unhovered")
+Style("focus_mask")
 add(ui_properties)
 add(position_properties)
 
@@ -888,13 +910,14 @@ add(position_properties)
 ##############################################################################
 # Control-flow statements.
 
-def PassParser(Parser):
+class PassParser(Parser):
 
     def __init__(self, name):
         super(PassParser, self).__init__(name)
 
     def parse(self, l, name):
-        return [ ast.Pass(lineno=l.number, col_offset=0) ]
+        return self.parse_exec("pass", l.number)
+
 
 PassParser("pass")
 
@@ -904,7 +927,6 @@ class DefaultParser(Parser):
     def __init__(self, name):
         super(DefaultParser, self).__init__(name)
 
-
     def parse(self, l, name):
 
         name = l.require(l.word)
@@ -915,6 +937,7 @@ class DefaultParser(Parser):
 
         return self.parse_exec(code, l.number)
 
+
 DefaultParser("default")
 
 
@@ -956,6 +979,7 @@ class UseParser(Parser):
 
         return self.parse_exec(code, lineno)
 
+
 UseParser("use")
 
 
@@ -1019,6 +1043,7 @@ class IfParser(Parser):
 
         return [ rv ]
 
+
 IfParser("if")
 
 
@@ -1084,15 +1109,16 @@ class ForParser(Parser):
             rv = self.parse_exec("%s = 0" % counter_name)
 
             rv.append(ast.For(
-                    target=pattern,
-                    iter=expression,
-                    body=children,
-                    orelse=[],
-                    lineno=lineno,
-                    col_offset=0))
+                target=pattern,
+                iter=expression,
+                body=children,
+                orelse=[],
+                lineno=lineno,
+                col_offset=0))
 
         return rv
 
+
 ForParser("for")
 
 
@@ -1118,6 +1144,7 @@ class PythonParser(Parser):
 
         return self.parse_exec(python_code, lineno)
 
+
 PythonParser("$", True)
 PythonParser("python", False)
 
@@ -1177,6 +1204,7 @@ class ScreenLangScreen(renpy.object.Object):
     predict = "False"
 
     parameters = None
+    location = None
 
     def __init__(self):
 
@@ -1196,10 +1224,10 @@ class ScreenLangScreen(renpy.object.Object):
         self.code = None
 
         # The variant of screen we're defining.
-        self.variant = "None" # expr.
+        self.variant = "None"  # expr.
 
         # Should we predict this screen?
-        self.predict = "None" # expr.
+        self.predict = "None"  # expr.
 
         # The parameters this screen takes.
         self.parameters = None
@@ -1209,7 +1237,7 @@ class ScreenLangScreen(renpy.object.Object):
             self.modal = "False"
             self.zorder = "0"
 
-    def define(self):
+    def define(self, location):
         """
         Defines a screen.
         """
@@ -1223,6 +1251,7 @@ class ScreenLangScreen(renpy.object.Object):
             variant=renpy.python.py_eval(self.variant),
             predict=renpy.python.py_eval(self.predict),
             parameters=self.parameters,
+            location=self.location,
             )
 
     def __call__(self, *args, **kwargs):
@@ -1239,7 +1268,6 @@ class ScreenLangScreen(renpy.object.Object):
         renpy.python.py_exec_bytecode(self.code.bytecode, locals=scope)
 
 
-
 class ScreenParser(Parser):
 
     def __init__(self):
@@ -1332,9 +1360,11 @@ class ScreenParser(Parser):
 
         return screen
 
+
 screen_parser = ScreenParser()
 screen_parser.add(all_statements)
 
+
 def parse_screen(l):
     """
     Parses the screen statement.
