@@ -1,4 +1,4 @@
-# Copyright 2004-2014 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,7 +19,9 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-import pygame
+from __future__ import print_function
+
+import pygame_sdl2 as pygame
 
 try:
     import xml.etree.ElementTree as etree
@@ -30,11 +32,31 @@ import renpy.display
 import renpy.text.ftfont as ftfont
 import renpy.text.textsupport as textsupport
 
-ftfont.init() #@UndefinedVariable
+ftfont.init()  # @UndefinedVariable
 
 WHITE = (255, 255, 255, 255)
 BLACK = (0, 0, 0, 255)
 
+
+def is_zerowidth(char):
+    if char == 0x200b:  # Zero-width space.
+        return True
+
+    if char == 0x200c:  # Zero-width non-joiner.
+        return True
+
+    if char == 0x200d:  # Zero-width joiner.
+        return True
+
+    if char == 0x2060:  # Word joiner.
+        return True
+
+    if char == 0xfeff:  # Zero width non-breaking space.
+        return True
+
+    return False
+
+
 class ImageFont(object):
 
     # ImageFonts are expected to have the following fields defined by
@@ -52,7 +74,6 @@ class ImageFont(object):
     # offsets - The x and y offsets of each character.
     # chars - A map from a character to the surface containing that character.
 
-
     def glyphs(self, s):
 
         rv = [ ]
@@ -61,18 +82,24 @@ class ImageFont(object):
             return rv
 
         for c in s:
-            g = textsupport.Glyph() #@UndefinedVariable
+            g = textsupport.Glyph()  # @UndefinedVariable
 
             g.character = ord(c)
-            g.ascent = self.height
+            g.ascent = self.baseline
             g.line_spacing = self.height
 
-            width = self.width.get(c, None)
-            if width is None:
-                raise Exception("Character {0!r} not found in image-based font.".format(c))
+            if not is_zerowidth(g.character):
 
-            g.width = self.width[c]
-            g.advance = self.advance[c]
+                width = self.width.get(c, None)
+                if width is None:
+                    raise Exception("Character {0!r} not found in image-based font.".format(c))
+
+                g.width = self.width[c]
+                g.advance = self.advance[c]
+
+            else:
+                g.width = 0
+                g.advance = 0
 
             rv.append(g)
 
@@ -92,6 +119,10 @@ class ImageFont(object):
             return
 
         for g in glyphs:
+
+            if not g.width:
+                continue
+
             c = unichr(g.character)
 
             cxo, cyo = self.offsets[c]
@@ -117,7 +148,8 @@ class SFont(ImageFont):
                  spacewidth,
                  default_kern,
                  kerns,
-                 charset):
+                 charset,
+                 baseline=None):
 
         super(SFont, self).__init__()
 
@@ -126,21 +158,26 @@ class SFont(ImageFont):
         self.default_kern = default_kern
         self.kerns = kerns
         self.charset = charset
+        self.baseline = baseline
 
     def load(self):
 
-        self.chars = { } # W0201
-        self.width = { } # W0201
-        self.advance = { } # W0201
-        self.offsets = { } # W0201
+        self.chars = { }  # W0201
+        self.width = { }  # W0201
+        self.advance = { }  # W0201
+        self.offsets = { }  # W0201
 
         # Load in the image.
         surf = renpy.display.im.Image(self.filename).load(unscaled=True)
 
         sw, sh = surf.get_size()
         height = sh
-        self.height = height # W0201
-        self.baseline = height # W0201
+        self.height = height  # W0201
+        if self.baseline is None:
+            self.baseline = height  # W0201
+        elif self.baseline < 0:
+            # Negative value is the distance from the bottom (vs top)
+            self.baseline = height + self.baseline  # W0201
 
         # Create space characters.
         self.chars[u' '] = renpy.display.pgrender.surface((self.spacewidth, height), True)
@@ -214,10 +251,10 @@ class MudgeFont(ImageFont):
 
     def load(self):
 
-        self.chars = { } # W0201
-        self.width = { } # W0201
-        self.advance = { } # W0201
-        self.offsets = { } # W0201
+        self.chars = { }  # W0201
+        self.width = { }  # W0201
+        self.advance = { }  # W0201
+        self.offsets = { }  # W0201
 
         # Load in the image.
         surf = renpy.display.im.Image(self.filename).load(unscaled=True)
@@ -250,8 +287,8 @@ class MudgeFont(ImageFont):
 
             height = max(height, h)
 
-        self.height = height # W0201
-        self.baseline = height # W0201
+        self.height = height  # W0201
+        self.baseline = height  # W0201
 
         # Create space characters.
         if u' ' not in self.chars:
@@ -272,7 +309,6 @@ class MudgeFont(ImageFont):
         self.offsets[u'\u200b'] = (0, 0)
 
 
-
 def parse_bmfont_line(l):
     w = ""
     line = [ ]
@@ -298,9 +334,10 @@ def parse_bmfont_line(l):
     if w:
         line.append(w)
 
-    map = dict(i.split("=", 1) for i in line[1:]) #@ReservedAssignment
+    map = dict(i.split("=", 1) for i in line[1:])  # @ReservedAssignment
     return line[0], map
 
+
 class BMFont(ImageFont):
 
     def __init__(self, filename):
@@ -310,12 +347,12 @@ class BMFont(ImageFont):
 
     def load(self):
 
-        self.chars = { } # W0201
-        self.width = { } # W0201
-        self.advance = { } # W0201
-        self.offsets = { } # W0201
-        self.kerns = { } # W0201
-        self.default_kern = 0 # W0201
+        self.chars = { }  # W0201
+        self.width = { }  # W0201
+        self.advance = { }  # W0201
+        self.offsets = { }  # W0201
+        self.kerns = { }  # W0201
+        self.default_kern = 0  # W0201
 
         pages = { }
 
@@ -325,8 +362,8 @@ class BMFont(ImageFont):
             kind, args = parse_bmfont_line(l)
 
             if kind == "common":
-                self.height = int(args["lineHeight"]) # W0201
-                self.baseline = int(args["base"]) # W0201
+                self.height = int(args["lineHeight"])  # W0201
+                self.baseline = int(args["base"])  # W0201
             elif kind == "page":
                 pages[int(args["id"])] = renpy.display.im.Image(args["file"]).load(unscaled=True)
             elif kind == "char":
@@ -347,6 +384,10 @@ class BMFont(ImageFont):
                 self.width[c] = w + xo
                 self.advance[c] = xadvance
                 self.offsets[c] = (xo, yo)
+            elif kind == "kerning":
+                first = unichr(int(args["first"]))
+                second = unichr(int(args["second"]))
+                self.kerns[first + second] = int(args["amount"])
 
         f.close()
 
@@ -356,17 +397,43 @@ class BMFont(ImageFont):
             self.advance[u'\u00a0'] = self.advance[u' ']
             self.offsets[u'\u00a0'] = self.offsets[u' ']
 
-
         self.chars[u'\u200b'] = renpy.display.pgrender.surface((0, self.height), True)
         self.width[u'\u200b'] = 0
         self.advance[u'\u200b'] = 0
         self.offsets[u'\u200b'] = (0, 0)
 
 
-def register_sfont(name=None, size=None, bold=False, italics=False, underline=False,
-                   filename=None, spacewidth=10, default_kern=0, kerns={},
-                   charset=u"!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~"):
+class ScaledImageFont(ImageFont):
+    """
+    Represents an imagefont scaled by a given factor.
+    """
 
+    def __init__(self, parent, factor):
+
+        def scale(n):
+            return int(round(n * factor))
+
+        self.height = scale(parent.height)
+        self.baseline = scale(parent.baseline)
+        self.default_kern = scale(parent.default_kern)
+
+        self.width = { k : scale(v) for k, v in parent.width.iteritems() }
+        self.advance = { k : scale(v) for k, v in parent.advance.iteritems() }
+        self.offsets = { k : (scale(v[0]), scale(v[1])) for k, v in parent.offsets.iteritems() }
+        self.kerns = { k : scale(v) for k, v in parent.kerns.iteritems() }
+
+        self.chars = { }
+
+        for k, v in parent.chars.iteritems():
+            w, h = v.get_size()
+            nw = scale(w)
+            nh = scale(h)
+            self.chars[k] = renpy.display.scale.smoothscale(v, (nw, nh))
+
+
+def register_sfont(name=None, size=None, bold=False, italics=False, underline=False,
+                   filename=None, spacewidth=10, baseline=None, default_kern=0, kerns={},
+                   charset=u"!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~"):
     """
     :doc: image_fonts
 
@@ -397,6 +464,13 @@ def register_sfont(name=None, size=None, bold=False, italics=False, underline=Fa
     `spacewidth`
         The width of a space character, an integer in pixels.
 
+    `baseline`
+        The distance from the top of the font to the baseline (the invisible
+        line letters sit on), an integer in pixels.  If this font is mixed with
+        other fonts, their baselines will be aligned.  Negative values indicate
+        distance from the bottom of the font instead, and ``None`` means the
+        baseline equals the height (i.e., is at the very bottom of the font).
+
     `default_kern`
         The default kern spacing between characters, in pixels.
 
@@ -404,7 +478,8 @@ def register_sfont(name=None, size=None, bold=False, italics=False, underline=Fa
         A map from two-character strings to the kern that should be used between
         those characters.
 
-    `charset` - The character set of the font. A string containing characters in
+    `charset`
+        The character set of the font. A string containing characters in
         the order in which they are found in the image. The default character
         set for a SFont is::
 
@@ -416,12 +491,12 @@ def register_sfont(name=None, size=None, bold=False, italics=False, underline=Fa
     if name is None or size is None or filename is None:
         raise Exception("When registering an SFont, the font name, font size, and filename are required.")
 
-    sf = SFont(filename, spacewidth, default_kern, kerns, charset)
+    sf = SFont(filename, spacewidth, default_kern, kerns, charset, baseline)
     image_fonts[(name, size, bold, italics)] = sf
 
-def register_mudgefont(name=None, size=None, bold=False, italics=False, underline=False,
-                   filename=None, xml=None, spacewidth=10, default_kern=0, kerns={}):
 
+def register_mudgefont(name=None, size=None, bold=False, italics=False, underline=False,
+                       filename=None, xml=None, spacewidth=10, default_kern=0, kerns={}):
     """
     :doc: image_fonts
 
@@ -451,7 +526,7 @@ def register_mudgefont(name=None, size=None, bold=False, italics=False, underlin
 
     `filename`
         The file containing the MudgeFont image, a string. The image is usually
-        a TGA file, but could be a PNG or other format that Ren'PY supports.
+        a TGA file, but could be a PNG or other format that Ren'Py supports.
 
     `xml`
         The xml file containing information generated by the MudgeFont tool.
@@ -473,9 +548,9 @@ def register_mudgefont(name=None, size=None, bold=False, italics=False, underlin
     mf = MudgeFont(filename, xml, spacewidth, default_kern, kerns)
     image_fonts[(name, size, bold, italics)] = mf
 
+
 def register_bmfont(name=None, size=None, bold=False, italics=False, underline=False,
                     filename=None):
-
     """
     :doc: image_fonts
 
@@ -515,9 +590,11 @@ def register_bmfont(name=None, size=None, bold=False, italics=False, underline=F
     bmf = BMFont(filename)
     image_fonts[(name, size, bold, italics)] = bmf
 
+
 # A map from face name to ftfont.FTFace
 face_cache = { }
 
+
 def load_face(fn):
 
     if fn in face_cache:
@@ -538,78 +615,111 @@ def load_face(fn):
         font_file = renpy.loader.load(fn)
     except IOError:
 
-        # Let's try to find the font on our own.
-        fonts = [ i.strip().lower() for i in fn.split(",") ]
+        if (not renpy.config.developer) or renpy.config.allow_sysfonts:
 
-        pygame.sysfont.initsysfonts()
+            # Let's try to find the font on our own.
+            fonts = [ i.strip().lower() for i in fn.split(",") ]
 
-        for v in pygame.sysfont.Sysfonts.itervalues():
-            if v is not None:
-                for _flags, ffn in v.iteritems():
-                    for i in fonts:
-                        if ffn.lower().endswith(i):
-                            font_file = file(ffn, "rb")
+            pygame.sysfont.initsysfonts()
+
+            for v in pygame.sysfont.Sysfonts.itervalues():
+                if v is not None:
+                    for _flags, ffn in v.iteritems():
+                        for i in fonts:
+                            if ffn.lower().endswith(i):
+                                font_file = file(ffn, "rb")
+                                break
+
+                        if font_file:
                             break
 
-                    if font_file:
-                        break
-
-            if font_file:
-                break
-
+                if font_file:
+                    break
 
     if font_file is None:
         raise Exception("Could not find font {0!r}.".format(orig_fn))
 
-    rv = ftfont.FTFace(font_file, index) #@UndefinedVariable
+    rv = ftfont.FTFace(font_file, index, orig_fn)  # @UndefinedVariable
 
     face_cache[orig_fn] = rv
 
     return rv
 
+
 # Caches of fonts.
 image_fonts = { }
+
+# A cache of scaled image fonts.
+scaled_image_fonts = { }
+
+# A cache of scaled faces.
 font_cache = { }
 
-def get_font(fn, size, bold, italics, outline, antialias, vertical):
+# The last_scale we last accessed fonts at. (Used to clear caches.)
+last_scale = 1.0
 
+
+def get_font(fn, size, bold, italics, outline, antialias, vertical, hinting, scale):
+
+    # If the scale changed, invalidate caches of scaled fonts.
+    global last_scale
+
+    if (scale != 1.0) and (scale != last_scale):
+        scaled_image_fonts.clear()
+        font_cache.clear()
+        last_scale = scale
+
+    # Perform replacement.
     t = (fn, bold, italics)
     fn, bold, italics = renpy.config.font_replacement_map.get(t, t)
 
-    rv = image_fonts.get((fn, size, bold, italics), None)
+    # Image fonts.
+    key = (fn, size, bold, italics)
+
+    rv = image_fonts.get(key, None)
     if rv is not None:
+
+        if scale != 1.0:
+            if key in scaled_image_fonts:
+                rv = scaled_image_fonts[key]
+            else:
+                rv = ScaledImageFont(rv, scale)
+                scaled_image_fonts[key] = rv
+
         return rv
 
-    if vertical:
-        key = (fn, size, bold, italics, outline, antialias, True)
-    else:
-        key = (fn, size, bold, italics, outline, antialias)
+    # Check for a cached TTF.
+    key = (fn, size, bold, italics, outline, antialias, vertical, hinting, scale)
 
     rv = font_cache.get(key, None)
     if rv is not None:
         return rv
 
-    # If we made it here, we need to load a ttf.
+    # Load a TTF.
     face = load_face(fn)
-
-    rv = ftfont.FTFont(face, size, bold, italics, outline, antialias, vertical) #@UndefinedVariable
+    rv = ftfont.FTFont(face, int(size * scale), bold, italics, outline, antialias, vertical, hinting)  # @UndefinedVariable
 
     font_cache[key] = rv
 
     return rv
 
+
 def free_memory():
     """
     Clears the font cache.
     """
 
+    scaled_image_fonts.clear()
     font_cache.clear()
-    face_cache.clear()
 
-def load_image_fonts():
+
+def load_fonts():
     for i in image_fonts.itervalues():
         i.load()
 
+    for i in renpy.config.preload_fonts:
+        load_face(i)
+
 
 class FontGroup(object):
     """
@@ -621,14 +731,9 @@ class FontGroup(object):
 
     def __init__(self):
 
-        # A list of font names we know of.
-        self.fonts = [ ]
-
-        # A map from character to the index of the font it's part of.
-        self.cache = { }
-
-        # A list of (index, start, end) tuples.
-        self.patterns = [ ]
+        # A map from character index to font name. None is used for
+        # the default font.
+        self.map = { }
 
     def add(self, font, start, end):
         """
@@ -638,7 +743,8 @@ class FontGroup(object):
 
         `start`
             The start of the range. This may be a single-character string, or
-            an integer giving a unicode code point.
+            an integer giving a unicode code point. If start is None, then the
+            font is used as the default.
 
         `end`
             The end of the range. This may be a single-character string, or an
@@ -651,6 +757,18 @@ class FontGroup(object):
         chained together.
         """
 
+        if start is None:
+
+            if isinstance(font, FontGroup):
+                for k, v in font.map.items():
+                    if k not in self.map:
+                        self.map[k] = v
+            else:
+                if None not in self.map:
+                    self.map[None] = font
+
+            return self
+
         if not isinstance(start, int):
             start = ord(start)
 
@@ -660,12 +778,9 @@ class FontGroup(object):
         if end < start:
             raise Exception("In FontGroup.add, the start of a character range must be before the end of the range.")
 
-        if font not in self.fonts:
-            self.fonts.append(font)
-
-        index = self.fonts.index(font)
-
-        self.patterns.append((index, start, end))
+        for i in range(start, end+1):
+            if i not in self.map:
+                self.map[i] = font
 
         return self
 
@@ -677,33 +792,27 @@ class FontGroup(object):
         mark = 0
         pos = 0
 
-        old_index = 0
-
-        cache = self.cache
+        old_font = None
 
         for c in s:
 
-            index = cache.get(c, None)
+            n = ord(c)
 
-            if index is None:
-                n = ord(c)
+            font = self.map.get(ord(c), None)
 
-                for index, start, end in self.patterns:
-                    if start <= n <= end:
-                        break
-                else:
+            if font is None:
+                font = self.map.get(None, None)
+
+                if font is None:
                     raise Exception("Character U+{0:04x} not found in FontGroup".format(n))
 
-                cache[c] = index
-
-            if index != old_index:
+            if font != old_font:
                 if pos:
-                    yield self.fonts[old_index], s[mark:pos]
+                    yield old_font, s[mark:pos]
 
-                old_index = index
+                old_font = font
                 mark = pos
 
             pos += 1
 
-        yield self.fonts[old_index], s[mark:]
-
+        yield font, s[mark:]
